import shutil
import urllib.request
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union
from zipfile import ZipFile

from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, Seconds


def download_and_unzip(
        target_dir: Pathlike = '.',
        force_download: Optional[bool] = False,
        url: Optional[str] = 'https://zenodo.org/record/3871592/files/MiniLibriMix.zip'
) -> None:
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    zip_path = target_dir / 'MiniLibriMix.zip'
    if force_download or not zip_path.is_file():
        urllib.request.urlretrieve(url, filename=zip_path)
    unzipped_dir = target_dir / 'MiniLibriMix'
    completed_detector = unzipped_dir / '.completed'
    if not completed_detector.is_file():
        shutil.rmtree(unzipped_dir, ignore_errors=True)
        with ZipFile(zip_path) as zf:
            zf.extractall(path=target_dir)
            completed_detector.touch()


def prepare_librimix(
        librimix_csv: Pathlike,
        output_dir: Optional[Pathlike] = None,
        with_precomputed_mixtures: bool = False,
        sampling_rate: int = 16000,
        min_segment_seconds: Seconds = 3.0
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    import pandas as pd
    assert Path(librimix_csv).is_file(), f'No such file: {librimix_csv}'
    df = pd.read_csv(librimix_csv)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    # First, create the audio manifest that specifies the pairs of source recordings
    # to be mixed together.
    audio_sources = RecordingSet.from_recordings(
        Recording(
            id=row['mixture_ID'],
            sources=[
                AudioSource(
                    type='file',
                    channels=[0],
                    source=row['source_1_path']
                ),
                AudioSource(
                    type='file',
                    channels=[1],
                    source=row['source_2_path']
                )
            ],
            sampling_rate=sampling_rate,
            num_samples=int(row['length']),
            duration=row['length'] / sampling_rate
        )
        for idx, row in df.iterrows()
        if row['length'] / sampling_rate > min_segment_seconds
    )
    supervision_sources = make_corresponding_supervisions(audio_sources)
    if output_dir is not None:
        audio_sources.to_json(output_dir / 'recordings_sources.json')
        supervision_sources.to_json(output_dir / 'supervisions_sources.json')

    manifests['sources'] = {
        'recordings': audio_sources,
        'supervisions': supervision_sources
    }

    # When requested, create an audio manifest for the pre-computed mixtures.
    # A different way of performing the mix would be using Lhotse's on-the-fly
    # overlaying of audio Cuts.
    if with_precomputed_mixtures:
        audio_mix = RecordingSet.from_recordings(
            Recording(
                id=row['mixture_ID'],
                sources=[
                    AudioSource(
                        type='file',
                        channels=[0],
                        source=row['mixture_path']
                    ),
                ],
                sampling_rate=sampling_rate,
                num_samples=int(row['length']),
                duration=row['length'] / sampling_rate
            )
            for idx, row in df.iterrows()
            if row['length'] / sampling_rate > min_segment_seconds
        )
        supervision_mix = make_corresponding_supervisions(audio_mix)
        if output_dir is not None:
            audio_mix.to_json(output_dir / 'recordings_mix.json')
            supervision_mix.to_json(output_dir / 'supervisions_mix.json')
        manifests['premixed'] = {
            'recordings': audio_mix,
            'supervisions': supervision_mix
        }

    # When the LibriMix CSV specifies noises, we create a separate RecordingSet for them,
    # so that we can extract their features and overlay them as Cuts later.
    if 'noise_path' in df:
        audio_noise = RecordingSet.from_recordings(
            Recording(
                id=row['mixture_ID'],
                sources=[
                    AudioSource(
                        type='file',
                        channels=[0],
                        source=row['noise_path']
                    ),
                ],
                sampling_rate=sampling_rate,
                num_samples=int(row['length']),
                duration=row['length'] / sampling_rate
            )
            for idx, row in df.iterrows()
            if row['length'] / sampling_rate > min_segment_seconds
        )
        supervision_noise = make_corresponding_supervisions(audio_noise)
        if output_dir is not None:
            audio_noise.to_json(output_dir / 'recordings_noise.json')
            supervision_noise.to_json(output_dir / 'supervisions_noise.json')
        manifests['noise'] = {
            'recordings': audio_noise,
            'supervisions': supervision_noise
        }

    return manifests


def make_corresponding_supervisions(audio: RecordingSet) -> SupervisionSet:
    """
    Prepare a supervision set - in this case it just describes
    which segments are available in the corpus, as the actual supervisions for
    speech separation come from the source recordings.
    """
    return SupervisionSet.from_segments(
        SupervisionSegment(
            id=f'{recording.id}-c{source.channels[0]}',
            recording_id=recording.id,
            start=0.0,
            duration=recording.duration,
            channel=source.channels[0],
        )
        for recording in audio
        for source in recording.sources
    )



from functools import reduce
from itertools import chain
from operator import add
from typing import Iterable, Optional, TypeVar

from lhotse.audio import Recording, RecordingSet
from lhotse.cut import Cut, CutSet, MixedCut
from lhotse.features import FeatureSet, Features
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, load_yaml

ManifestItem = TypeVar('ManifestItem', Recording, SupervisionSegment, Features, Cut, MixedCut)
Manifest = TypeVar('Manifest', RecordingSet, SupervisionSet, FeatureSet, CutSet)


def combine(*manifests: Manifest) -> Manifest:
    """Combine multiple manifests of the same type into one."""
    return reduce(add, manifests)


def to_manifest(items: Iterable[ManifestItem]) -> Optional[Manifest]:
    """
    Take an iterable of data types in Lhotse such as Recording, SupervisonSegment or Cut, and create the manifest of the
    corresponding type. When the iterable is empty, returns None.
    """
    items = iter(items)
    try:
        first_item = next(items)
    except StopIteration:
        return None
    items = chain([first_item], items)

    if isinstance(first_item, Recording):
        return RecordingSet.from_recordings(items)
    if isinstance(first_item, SupervisionSegment):
        return SupervisionSet.from_segments(items)
    if isinstance(first_item, (Cut, MixedCut)):
        return CutSet.from_cuts(items)
    if isinstance(first_item, Features):
        raise ValueError("FeatureSet generic construction from iterable is not possible, as the config information "
                         "would have been lost. Call FeatureSet.from_features() directly instead.")

    raise ValueError(f"Unknown type of manifest item: {first_item}")


def load_manifest(path: Pathlike) -> Manifest:
    """Generic utility for reading an arbitrary manifest."""
    raw_data = load_yaml(path)
    data_set = None
    for manifest_type in [RecordingSet, SupervisionSet, FeatureSet, CutSet]:
        try:
            data_set = manifest_type.from_dicts(raw_data)
        except Exception:
            pass
    if data_set is None:
        raise ValueError(f'Unknown type of manifest: {path}')
    return data_set



from .audio import RecordingSet, Recording, AudioSource
from .augmentation import WavAugmenter
from .cut import CutSet, Cut
from .features import *
from .kaldi import load_kaldi_data_dir
from .manipulation import load_manifest
from .supervision import SupervisionSet, SupervisionSegment



from typing import Optional

import numpy as np

from lhotse.features.base import FeatureExtractor
from lhotse.utils import Seconds, Decibels


class FeatureMixer:
    """
    Utility class to mix multiple feature matrices into a single one.
    It should be instantiated separately for each mixing session (i.e. each ``MixedCut``
    will create a separate ``FeatureMixer`` to mix its tracks).
    It is initialized with a numpy array of features (typically float32)
    that represents the "reference" signal for the mix.
    Other signals can be mixed to it with different time offsets and SNRs using the
    ``add_to_mix`` method.
    The time offset is relative to the start of the reference signal
    (only positive values are supported).
    The SNR is relative to the energy of the signal used to initialize the ``FeatureMixer``.

    It relies on the ``FeatureExtractor`` to have defined ``mix`` and ``compute_energy`` methods,
    so that the ``FeatureMixer`` knows how to scale and add two feature matrices together.
    """

    def __init__(
            self,
            feature_extractor: FeatureExtractor,
            base_feats: np.ndarray,
            frame_shift: Seconds,
            padding_value: float = -1000.0
    ):
        """
        :param feature_extractor: The ``FeatureExtractor`` instance that specifies how to mix the features.
        :param base_feats: The features used to initialize the ``FeatureMixer`` are a point of reference
            in terms of energy and offset for all features mixed into them.
        :param frame_shift: Required to correctly compute offset and padding during the mix.
        :param padding_value: The value used to pad the shorter features during the mix.
            This value is adequate only for log space features. For non-log space features,
            e.g. energies, use either 0 or a small positive value like 1e-5.
        """
        self.feature_extractor = feature_extractor
        self.tracks = [base_feats]
        self.gains = []
        # Keep a pre-computed energy value of the features that we initialize the Mixer with;
        # it is required to compute gain ratios that satisfy SNR during the mix.
        self.frame_shift = frame_shift
        self.reference_energy = feature_extractor.compute_energy(base_feats)
        assert self.reference_energy > 0.0, \
            f"To perform mix, energy must be non-zero and non-negative (got {self.reference_energy})"
        self.padding_value = padding_value
        self.dtype = self.tracks[0].dtype

    @property
    def num_features(self):
        return self.tracks[0].shape[1]

    @property
    def unmixed_feats(self) -> np.ndarray:
        """
        Return a numpy ndarray with the shape (num_tracks, num_frames, num_features), where each track's
        feature matrix is padded and scaled adequately to the offsets and SNR used in ``add_to_mix`` call.
        """
        return np.stack(self.tracks)

    @property
    def mixed_feats(self) -> np.ndarray:
        """
        Return a numpy ndarray with the shape (num_frames, num_features) - a mono mixed feature matrix
        of the tracks supplied with ``add_to_mix`` calls.
        """
        result = self.tracks[0]
        for feats_to_add, gain in zip(self.tracks[1:], self.gains):
            result = self.feature_extractor.mix(
                features_a=result,
                features_b=feats_to_add,
                energy_scaling_factor_b=gain
            )
        return result

    def add_to_mix(
            self,
            feats: np.ndarray,
            snr: Optional[Decibels] = None,
            offset: Seconds = 0.0
    ):
        """
        Add feature matrix of a new track into the mix.
        :param feats: A 2D feature matrix to be mixed in.
        :param snr: Signal-to-noise ratio, assuming ``feats`` represents noise (positive SNR - lower ``feats`` energy,
        negative SNR - higher ``feats`` energy)
        :param offset: How many seconds to shift ``feats`` in time. For mixing, the signal will be padded before
        the start with low energy values.
        """
        assert offset >= 0.0, "Negative offset in mixing is not supported."

        reference_feats = self.tracks[0]
        num_frames_offset = round(offset / self.frame_shift)
        current_num_frames = reference_feats.shape[0]
        incoming_num_frames = feats.shape[0] + num_frames_offset
        mix_num_frames = max(current_num_frames, incoming_num_frames)

        feats_to_add = feats

        # When the existing frames are less than what we anticipate after the mix,
        # we need to pad after the end of the existing features mixed so far.
        if current_num_frames < mix_num_frames:
            for idx in range(len(self.tracks)):
                padded_track = np.vstack([
                    self.tracks[idx],
                    self.padding_value * np.ones(
                        (mix_num_frames - current_num_frames, self.num_features),
                        dtype=self.dtype
                    )
                ])
                self.tracks[idx] = padded_track

        # When there is an offset, we need to pad before the start of the features we're adding.
        if offset > 0:
            feats_to_add = np.vstack([
                self.padding_value * np.ones(
                    (num_frames_offset, self.num_features),
                    dtype=self.dtype
                ),
                feats_to_add
            ])

        # When the features we're mixing in are shorter that the anticipated mix length,
        # we need to pad after their end.
        # Note: we're doing that inefficiently, as we potentially re-allocate numpy arrays twice,
        # during this padding and the offset padding before. If that's a bottleneck, we'll optimize.
        if incoming_num_frames < mix_num_frames:
            feats_to_add = np.vstack([
                feats_to_add,
                self.padding_value * np.ones(
                    (mix_num_frames - incoming_num_frames, self.num_features),
                    dtype=self.dtype
                )
            ])

        # When SNR is requested, find what gain is needed to satisfy the SNR
        gain = 1.0
        if snr is not None:
            # Compute the added signal energy before it was padded
            added_feats_energy = self.feature_extractor.compute_energy(feats)
            assert added_feats_energy > 0.0, \
                f"To perform mix, energy must be non-zero and non-negative (got {self.reference_energy})"
            target_energy = self.reference_energy * (10.0 ** (-snr / 10))
            gain = target_energy / added_feats_energy

        self.tracks.append(feats_to_add)
        self.gains.append(gain)



from math import isclose
import pytest

import torch

from lhotse.utils import time_diff_to_num_frames
from lhotse.cut import CutSet
from lhotse.dataset.vad import VadDataset


@pytest.fixture
def cut_set():
    return CutSet.from_json('test/fixtures/ami/cuts.json')


def test_vad_dataset(cut_set):
    duration = 5.0
    feature_threhold = -8.0

    cuts = list(cut_set)
    assert isclose(cuts[0].duration, 6.0)
    assert len(cuts[0].supervisions) == 2

    v1_start = time_diff_to_num_frames(cuts[0].supervisions[0].start, frame_length=0, frame_shift=0.01)
    v1_end = time_diff_to_num_frames(cuts[0].supervisions[0].end, frame_length=0, frame_shift=0.01)
    v2_start = time_diff_to_num_frames(cuts[0].supervisions[1].start, frame_length=0, frame_shift=0.01)
    v2_end = time_diff_to_num_frames(cuts[0].supervisions[1].end, frame_length=0, frame_shift=0.01)
    end = time_diff_to_num_frames(duration, frame_length=0, frame_shift=0.01)

    dataset = VadDataset(cut_set.cut_into_windows(duration=duration))
    example = dataset[0]
    assert isclose(float(torch.mean(example['is_voice'][0:v1_start])), 0)
    assert isclose(float(torch.mean(example['is_voice'][v1_start:v1_end])), 1)
    assert isclose(float(torch.mean(example['is_voice'][v1_end:v2_start])), 0)
    assert isclose(float(torch.mean(example['is_voice'][v2_start:v2_end])), 1)
    assert isclose(float(torch.mean(example['is_voice'][v2_end:end])), 0)

    assert float(torch.mean(example['features'][0:v1_start])) < feature_threhold
    assert float(torch.mean(example['features'][v1_start:v1_end])) > feature_threhold
    assert float(torch.mean(example['features'][v1_end:v2_start])) < feature_threhold
    assert float(torch.mean(example['features'][v2_start:v2_end])) > feature_threhold
    assert float(torch.mean(example['features'][v2_end:end])) < feature_threhold



import logging
import re
import shutil
import tarfile
import urllib.request
from collections import defaultdict
from pathlib import Path
from typing import Dict, NamedTuple, Optional, Tuple, Union

import torchaudio

from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike

dataset_parts_full = ('dev-clean', 'dev-other', 'test-clean', 'test-other',
                      'train-clean-100', 'train-clean-360', 'train-other-500')
dataset_parts_mini = ('dev-clean-2', 'train-clean-5')


def download_and_untar(
        target_dir: Pathlike = '.',
        dataset_parts: Optional[Tuple[str]] = dataset_parts_mini,
        force_download: Optional[bool] = False,
        base_url: Optional[str] = 'http://www.openslr.org/resources'
) -> None:
    """
    Downdload and untar the dataset, supporting both LibriSpeech and MiniLibrispeech

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: dataset part name, e.g. 'train-clean-100', 'train-clean-5', 'dev-clean'
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    """

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    for part in dataset_parts:
        if part in dataset_parts_full:
            url = f'{base_url}/12'
        elif part in dataset_parts_mini:
            url = f'{base_url}/31'
        else:
            logging.warning(f'Invalid dataset part name: {part}')
            continue
        tar_name = f'{part}.tar.gz'
        tar_path = target_dir / tar_name
        if force_download or not tar_path.is_file():
            urllib.request.urlretrieve(f'{url}/{tar_name}', filename=tar_path)
        part_dir = target_dir / f'LibriSpeech/{part}'
        completed_detector = part_dir / '.completed'
        if not completed_detector.is_file():
            shutil.rmtree(part_dir, ignore_errors=True)
            with tarfile.open(tar_path) as tar:
                tar.extractall(path=target_dir)
                completed_detector.touch()


class LibriSpeechMetaData(NamedTuple):
    audio_path: Pathlike
    audio_info: torchaudio.sox_signalinfo_t
    text: str


def prepare_librispeech(
        corpus_dir: Pathlike,
        dataset_parts: Optional[Tuple[str]] = dataset_parts_mini,
        output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions

    :param corpus_dir: Pathlike, the path of the data dir.
    :param dataset_parts: dataset part name, e.g. 'train-clean-100', 'train-clean-5', 'dev-clean'
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f'No such directory: {corpus_dir}'
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    manifests = defaultdict(dict)
    for part in dataset_parts:
        # Generate a mapping: utt_id -> (audio_path, audio_info, text)
        metadata = {}
        part_path = corpus_dir / part
        for trans_path in part_path.rglob('*.txt'):
            with open(trans_path) as f:
                for line in f:
                    idx, text = line.split(maxsplit=1)
                    audio_path = part_path / Path(idx.replace('-', '/')).parent / f'{idx}.flac'
                    if audio_path.is_file():
                        # info[0]: info of the raw audio (e.g. channel number, sample rate, duration ... )
                        # info[1]: info about the encoding (e.g. FLAC/ALAW/ULAW ...)
                        info = torchaudio.info(str(audio_path))
                        metadata[idx] = LibriSpeechMetaData(audio_path=audio_path, audio_info=info[0], text=text)
                    else:
                        logging.warning(f'No such file: {audio_path}')

        # Audio
        audio = RecordingSet.from_recordings(
            Recording(
                id=idx,
                sources=[
                    AudioSource(
                        type='file',
                        channels=[0],
                        source=str(metadata[idx].audio_path)
                    )
                ],
                sampling_rate=int(metadata[idx].audio_info.rate),
                num_samples=metadata[idx].audio_info.length,
                duration=(metadata[idx].audio_info.length / metadata[idx].audio_info.rate)
            )
            for idx in metadata
        )

        # Supervision
        supervision = SupervisionSet.from_segments(
            SupervisionSegment(
                id=idx,
                recording_id=idx,
                start=0.0,
                duration=audio.recordings[idx].duration,
                channel=0,
                language='English',
                speaker=re.sub(r'-.*', r'', idx),
                text=metadata[idx].text.strip()
            )
            for idx in audio.recordings
        )

        if output_dir is not None:
            supervision.to_json(output_dir / f'supervisions_{part}.json')
            audio.to_json(output_dir / f'recordings_{part}.json')

        manifests[part] = {
            'recordings': audio,
            'supervisions': supervision
        }

    return manifests



from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import nullcontext as no_executor
from tempfile import TemporaryDirectory
from unittest.mock import Mock

import pytest

from lhotse import Recording, Cut, Fbank, CutSet
from lhotse.audio import AudioSource
from lhotse.cut import MixedCut
from lhotse.features.io import LilcomFilesWriter


@pytest.fixture
def recording():
    return Recording(
        id='rec',
        sources=[AudioSource(type='file', channels=[0, 1], source='test/fixtures/stereo.wav')],
        sampling_rate=8000,
        num_samples=8000,
        duration=1.0
    )


@pytest.fixture
def cut(recording):
    return Cut(id='cut', start=0, duration=1.0, channel=0, recording=recording)


def test_extract_features(cut):
    extractor = Fbank()
    arr = cut.compute_features(extractor=extractor)
    assert arr.shape[0] == 100
    assert arr.shape[1] == extractor.feature_dim(cut.sampling_rate)


def test_extract_and_store_features(cut):
    extractor = Fbank()
    with TemporaryDirectory() as tmpdir, LilcomFilesWriter(tmpdir) as storage:
        cut_with_feats = cut.compute_and_store_features(extractor=extractor, storage=storage)
        arr = cut_with_feats.load_features()
    assert arr.shape[0] == 100
    assert arr.shape[1] == extractor.feature_dim(cut.sampling_rate)


@pytest.mark.parametrize('mix_eagerly', [False, True])
def test_extract_and_store_features_from_mixed_cut(cut, mix_eagerly):
    mixed_cut = cut.append(cut)
    extractor = Fbank()
    with TemporaryDirectory() as tmpdir, LilcomFilesWriter(tmpdir) as storage:
        cut_with_feats = mixed_cut.compute_and_store_features(
            extractor=extractor,
            storage=storage,
            mix_eagerly=mix_eagerly
        )
        arr = cut_with_feats.load_features()
    assert arr.shape[0] == 200
    assert arr.shape[1] == extractor.feature_dim(mixed_cut.sampling_rate)


@pytest.fixture
def cut_set(cut):
    # The padding tests if feature extraction works correctly with a PaddingCut
    return CutSet.from_cuts([cut, cut.append(cut).pad(3.0)])


# The lines below try to import Dask (a distributed computing library for Python)
# so that we can test that parallel feature extraction through the "executor"
# interface works correctly in this case. Dask in not a requirement or a dependency
# of Lhotse, so we make the tests with it optional as well.

try:
    import distributed
except:
    distributed = Mock()


def is_dask_availabe():
    try:
        import dask
        import distributed
        return True
    except:
        return False


@pytest.mark.parametrize(
    'mix_eagerly', [False, True]
)
@pytest.mark.parametrize(
    'executor', [
        None,
        ThreadPoolExecutor,
        pytest.param(ProcessPoolExecutor, marks=pytest.mark.skip(reason='Hangs CI but otherwise seems to work...')),
        pytest.param(distributed.Client, marks=pytest.mark.skipif(not is_dask_availabe(), reason='Requires Dask'))
    ]
)
def test_extract_and_store_features_from_cut_set(cut_set, executor, mix_eagerly):
    extractor = Fbank()
    with TemporaryDirectory() as tmpdir, LilcomFilesWriter(tmpdir) as storage:
        with executor() if executor is not None else no_executor() as ex:
            cut_set_with_feats = cut_set.compute_and_store_features(
                extractor=extractor,
                storage=storage,
                mix_eagerly=mix_eagerly,
                executor=ex
            )

        # The same number of cuts
        assert len(cut_set_with_feats) == 2

        for orig_cut, feat_cut in zip(cut_set, cut_set_with_feats):
            # The ID is retained
            assert orig_cut.id == feat_cut.id
            # Features were attached
            assert feat_cut.has_features
            # Recording is retained unless mixing a MixedCut eagerly
            should_have_recording = not (mix_eagerly and isinstance(orig_cut, MixedCut))
            assert feat_cut.has_recording == should_have_recording

        cuts = list(cut_set_with_feats)

        arr = cuts[0].load_features()
        assert arr.shape[0] == 100
        assert arr.shape[1] == extractor.feature_dim(cuts[0].sampling_rate)

        arr = cuts[1].load_features()
        assert arr.shape[0] == 300
        assert arr.shape[1] == extractor.feature_dim(cuts[0].sampling_rate)



import pytest

from lhotse.cut import make_windowed_cuts_from_features
from lhotse.features import FeatureSet, Features


def features(rec_id, start, duration):
    """Helper method for fixture readability (specify only relevant attributes)."""
    return Features(recording_id=rec_id, channels=0, start=start, duration=duration, sampling_rate=16000,
                    type='irrelevant', num_frames=round(duration / 0.01), num_features=23,
                    storage_type='irrelevant', storage_path='irrelevant', storage_key='irrelevant')


@pytest.fixture
def feature_set():
    return FeatureSet(
        features=[
            features('rec-1', 0.0, 600.0),
            features('rec-2', 0.0, 357.0)
        ]
    )


# noinspection PyMethodMayBeStatic
class TestMakeWindowedCutsFromFeatures:
    def test_full_shift_no_shorter_cuts(self, feature_set):
        cut_set = make_windowed_cuts_from_features(
            feature_set=feature_set,
            cut_duration=5.0
        )

        assert len(cut_set) == 191
        assert len([c for c in cut_set if c.recording_id == 'rec-1']) == 120
        assert len([c for c in cut_set if c.recording_id == 'rec-2']) == 71

        assert all(c.duration == 5.0 for c in cut_set)

    def test_full_shift_with_shorter_cuts(self, feature_set):
        cut_set = make_windowed_cuts_from_features(
            feature_set=feature_set,
            cut_duration=5.0,
            keep_shorter_windows=True
        )

        assert len(cut_set) == 192
        assert len([c for c in cut_set if c.recording_id == 'rec-1']) == 120
        assert len([c for c in cut_set if c.recording_id == 'rec-2']) == 72

        assert not all(c.duration == 5.0 for c in cut_set)

    def test_half_shift_no_shorter_cuts(self, feature_set):
        cut_set = make_windowed_cuts_from_features(
            feature_set=feature_set,
            cut_duration=5.0,
            cut_shift=2.5
        )

        assert len(cut_set) == 380
        # below, the last window is only 2.5s duration
        assert len([c for c in cut_set if c.recording_id == 'rec-1']) == 239
        # below, the last two windows are 2.0s and 4.5s duration
        assert len([c for c in cut_set if c.recording_id == 'rec-2']) == 141

        assert all(c.duration == 5.0 for c in cut_set)

    def test_half_shift_with_shorter_cuts(self, feature_set):
        cut_set = make_windowed_cuts_from_features(
            feature_set=feature_set,
            cut_duration=5.0,
            cut_shift=2.5,
            keep_shorter_windows=True
        )

        assert len(cut_set) == 383
        assert len([c for c in cut_set if c.recording_id == 'rec-1']) == 240
        assert len([c for c in cut_set if c.recording_id == 'rec-2']) == 143

        assert not all(c.duration == 5.0 for c in cut_set)



from dataclasses import dataclass

import numpy as np
import torchaudio

from lhotse.features.base import TorchaudioFeatureExtractor, register_extractor
from lhotse.utils import Seconds, EPSILON


@dataclass
class FbankConfig:
    # Spectogram-related part
    dither: float = 0.0
    window_type: str = "povey"
    # Note that frame_length and frame_shift will be converted to milliseconds before torchaudio/Kaldi sees them
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    remove_dc_offset: bool = True
    round_to_power_of_two: bool = True
    energy_floor: float = 1e-10
    min_duration: float = 0.0
    preemphasis_coefficient: float = 0.97
    raw_energy: bool = True

    # Fbank-related part
    low_freq: float = 20.0
    high_freq: float = -400.0
    num_mel_bins: int = 40
    use_energy: bool = False
    vtln_low: float = 100.0
    vtln_high: float = -500.0
    vtln_warp: float = 1.0


@register_extractor
class Fbank(TorchaudioFeatureExtractor):
    """Log Mel energy filter bank feature extractor based on ``torchaudio.compliance.kaldi.fbank`` function."""
    name = 'fbank'
    config_type = FbankConfig
    feature_fn = staticmethod(torchaudio.compliance.kaldi.fbank)

    def feature_dim(self, sampling_rate: int) -> int:
        return self.config.num_mel_bins

    @staticmethod
    def mix(features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float) -> np.ndarray:
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b)
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(np.exp(features)))



from .ami import prepare_ami
from .broadcast_news import prepare_broadcast_news
from .heroico import prepare_heroico
from .librimix import prepare_librimix
from .librispeech import prepare_librispeech
from .ljspeech import prepare_ljspeech
from .switchboard import prepare_switchboard
from .tedlium import prepare_tedlium



import random
import warnings
from dataclasses import dataclass, field
from functools import reduce
from math import ceil, floor
from typing import Any, Callable, Dict, FrozenSet, Iterable, List, Optional, Sequence, Union

import numpy as np
from cytoolz import sliding_window
from cytoolz.itertoolz import groupby

from lhotse import WavAugmenter
from lhotse.audio import AudioMixer, Recording, RecordingSet
from lhotse.features import FeatureExtractor, FeatureMixer, FeatureSet, Features, create_default_feature_extractor
from lhotse.features.io import FeaturesWriter
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import (Decibels, EPSILON, JsonMixin, Pathlike, Seconds, TimeSpan, YamlMixin, asdict_nonull, fastcopy,
                          overlaps, overspans, split_sequence, uuid4)

# One of the design principles for Cuts is a maximally "lazy" implementation, e.g. when mixing Cuts,
# we'd rather sum the feature matrices only after somebody actually calls "load_features". It helps to avoid
# an excessive storage size for data augmented in various ways.


# Helper "typedef" to artbitrary Cut type as they do not share a common base class.
# The class names are strings here so that the Python interpreter resolves them after parsing the whole file.
AnyCut = Union['Cut', 'MixedCut', 'PaddingCut']


# noinspection PyTypeChecker,PyUnresolvedReferences
class CutUtilsMixin:
    """
    A mixin class for cuts which contains all the methods that share common implementations.

    Note: Ideally, this would've been an abstract base class specifying the common interface,
    but ABC's do not mix well with dataclasses in Python. It is possible we'll ditch the dataclass
    for cuts in the future and make this an ABC instead.
    """

    def mix(self, other: AnyCut, offset_other_by: Seconds = 0.0, snr: Optional[Decibels] = None) -> 'MixedCut':
        """Refer to mix() documentation."""
        return mix(self, other, offset=offset_other_by, snr=snr)

    def append(self, other: AnyCut, snr: Optional[Decibels] = None) -> 'MixedCut':
        """
        Append the ``other`` Cut after the current Cut. Conceptually the same as ``mix`` but with an offset
        matching the current cuts length. Optionally scale down (positive SNR) or scale up (negative SNR)
        the ``other`` cut.
        Returns a MixedCut, which only keeps the information about the mix; actual mixing is performed
        during the call to ``load_features``.
        """
        return mix(self, other, offset=self.duration, snr=snr)

    def compute_features(
            self,
            extractor: FeatureExtractor,
            augmenter: Optional[WavAugmenter] = None,
    ) -> np.ndarray:
        """
        Compute the features from this cut. This cut has to be able to load audio.

        :param extractor: a ``FeatureExtractor`` instance used to compute the features.
        :param augmenter: optional ``WavAugmenter`` instance for audio augmentation.
        :return: a numpy ndarray with the computed features.
        """
        samples = self.load_audio()
        if augmenter is not None:
            samples = augmenter.apply(samples)
        return extractor.extract(samples, self.sampling_rate)

    def plot_audio(self):
        """
        Display a plot of the waveform. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt
        samples = self.load_audio().squeeze()
        fig, ax = plt.subplots()
        ax.plot(np.linspace(0, self.duration, len(samples)), samples)
        for supervision in self.supervisions:
            supervision = supervision.trim(self.duration)
            ax.axvspan(supervision.start, supervision.end, color='green', alpha=0.1)
        return ax

    def play_audio(self):
        """
        Display a Jupyter widget that allows to listen to the waveform.
        Works only in Jupyter notebook/lab or similar (e.g. Colab).
        """
        from IPython.display import Audio
        samples = self.load_audio().squeeze()
        return Audio(samples, rate=self.sampling_rate)

    def plot_features(self):
        """
        Display the feature matrix as an image. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt
        features = np.flip(self.load_features().transpose(1, 0), 0)
        return plt.matshow(features)

    def speakers_feature_mask(
            self,
            min_speaker_dim: Optional[int] = None,
            speaker_to_idx_map: Optional[Dict[str, int]] = None
    ) -> np.ndarray:
        """
        Return a matrix of per-speaker activity in a cut. The matrix shape is (num_speakers, num_frames),
        and its values are 0 for nonspeech **frames** and 1 for speech **frames** for each respective speaker.

        This is somewhat inspired by the TS-VAD setup: https://arxiv.org/abs/2005.07272

        :param min_speaker_dim: optional int, when specified it will enforce that the matrix shape is at least
            that value (useful for datasets like CHiME 6 where the number of speakers is always 4, but some cuts
            might have less speakers than that).
        :param speaker_to_idx_map: optional dict mapping speaker names (strings) to their global indices (ints).
            Useful when you want to preserve the order of the speakers (e.g. speaker XYZ is always mapped to index 2)
        """
        assert self.has_features, f"No features available. " \
                                  f"Can't compute supervisions feature mask for cut with ID: {self.id}."
        if speaker_to_idx_map is None:
            speaker_to_idx_map = {spk: idx for idx, spk in enumerate(sorted(set(s.speaker for s in self.supervisions)))}
        num_speakers = len(speaker_to_idx_map)
        if min_speaker_dim is not None:
            num_speakers = min(min_speaker_dim, num_speakers)
        mask = np.zeros((num_speakers, self.num_frames))
        for supervision in self.supervisions:
            speaker_idx = speaker_to_idx_map[supervision.speaker]
            st = round(supervision.start / self.frame_shift) if supervision.start > 0 else 0
            et = round(supervision.end / self.frame_shift) if supervision.end < self.duration else self.num_frames
            mask[speaker_idx, st:et] = 1
        return mask

    def speakers_audio_mask(
            self,
            min_speaker_dim: Optional[int] = None,
            speaker_to_idx_map: Optional[Dict[str, int]] = None
    ) -> np.ndarray:
        """
        Return a matrix of per-speaker activity in a cut. The matrix shape is (num_speakers, num_samples),
        and its values are 0 for nonspeech **samples** and 1 for speech **samples** for each respective speaker.

        This is somewhat inspired by the TS-VAD setup: https://arxiv.org/abs/2005.07272

        :param min_speaker_dim: optional int, when specified it will enforce that the matrix shape is at least
            that value (useful for datasets like CHiME 6 where the number of speakers is always 4, but some cuts
            might have less speakers than that).
        :param speaker_to_idx_map: optional dict mapping speaker names (strings) to their global indices (ints).
            Useful when you want to preserve the order of the speakers (e.g. speaker XYZ is always mapped to index 2)
        """
        assert self.has_recording, f"No recording available. " \
                                   f"Can't compute supervisions audio mask for cut with ID: {self.id}."
        if speaker_to_idx_map is None:
            speaker_to_idx_map = {spk: idx for idx, spk in enumerate(sorted(set(s.speaker for s in self.supervisions)))}
        num_speakers = len(speaker_to_idx_map)
        if min_speaker_dim is not None:
            num_speakers = min(min_speaker_dim, num_speakers)
        mask = np.zeros((num_speakers, self.num_samples))
        for supervision in self.supervisions:
            speaker_idx = speaker_to_idx_map[supervision.speaker]
            st = round(supervision.start * self.sampling_rate) if supervision.start > 0 else 0
            et = (
                round(supervision.end * self.sampling_rate)
                if supervision.end < self.duration
                else self.duration * self.sampling_rate
            )
            mask[speaker_idx, st:et] = 1
        return mask

    def supervisions_feature_mask(self) -> np.ndarray:
        """
        Return a 1D numpy array with value 1 for **frames** covered by at least one supervision,
        and 0 for **frames** not covered by any supervision.
        """
        assert self.has_features, f"No features available. " \
                                  f"Can't compute supervisions feature mask for cut with ID: {self.id}."
        mask = np.zeros(self.num_frames, dtype=np.float32)
        for supervision in self.supervisions:
            st = round(supervision.start / self.frame_shift) if supervision.start > 0 else 0
            et = round(supervision.end / self.frame_shift) if supervision.end < self.duration else self.num_frames
            mask[st:et] = 1.0
        return mask

    def supervisions_audio_mask(self) -> np.ndarray:
        """
        Return a 1D numpy array with value 1 for **samples** covered by at least one supervision,
        and 0 for **samples** not covered by any supervision.
        """
        assert self.has_recording, f"No recording available. " \
                                   f"Can't compute supervisions audio mask for cut with ID: {self.id}."
        mask = np.zeros(self.num_samples, dtype=np.float32)
        for supervision in self.supervisions:
            st = round(supervision.start * self.sampling_rate) if supervision.start > 0 else 0
            et = (
                round(supervision.end * self.sampling_rate)
                if supervision.end < self.duration
                else self.duration * self.sampling_rate
            )
            mask[st:et] = 1.0
        return mask

    def with_id(self, id_: str) -> AnyCut:
        """Return a copy of the Cut with a new ID."""
        return fastcopy(self, id=id_)


@dataclass
class Cut(CutUtilsMixin):
    """
    A Cut is a single "segment" that we'll train on. It contains the features corresponding to
    a piece of a recording, with zero or more SupervisionSegments.

    The SupervisionSegments indicate which time spans of the Cut contain some kind of supervision information:
    e.g. transcript, speaker, language, etc. The regions without a corresponding SupervisionSegment may
    contain anything - usually we assume it's either silence or some kind of noise.

    Note: The SupervisionSegment time boundaries are relative to the beginning of the cut.
    E.g. if the underlying Recording starts at 0s (always true), the Cut starts at 100s,
    and the SupervisionSegment starts at 3s, it means that in the Recording the supervision actually started at 103s.
    In some cases, the supervision might have a negative start, or a duration exceeding the duration of the Cut;
    this means that the supervision in the recording extends beyond the Cut.
    """
    id: str

    # Begin and duration are needed to specify which chunk of features/recording to load.
    start: Seconds
    duration: Seconds
    channel: int

    # Supervisions that will be used as targets for model training later on. They don't have to cover the whole
    # cut duration. They also might overlap.
    supervisions: List[SupervisionSegment] = field(default_factory=list)

    # The features can span longer than the actual cut - the Features object "knows" its start and end time
    # within the underlying recording. We can expect the interval [begin, begin + duration] to be a subset of the
    # interval represented in features.
    features: Optional[Features] = None

    # For the cases that the model was trained by raw audio instead of features
    recording: Optional[Recording] = None

    @property
    def recording_id(self) -> str:
        return self.features.recording_id if self.has_features else self.recording.id

    @property
    def end(self) -> Seconds:
        return round(self.start + self.duration, ndigits=3)

    @property
    def has_features(self) -> bool:
        return self.features is not None

    @property
    def has_recording(self) -> bool:
        return self.recording is not None

    @property
    def frame_shift(self) -> Optional[Seconds]:
        return self.features.frame_shift if self.has_features else None

    @property
    def num_frames(self) -> Optional[int]:
        return round(self.duration / self.frame_shift) if self.has_features else None

    @property
    def num_samples(self) -> Optional[int]:
        return round(self.duration * self.sampling_rate) if self.has_recording else None

    @property
    def num_features(self) -> Optional[int]:
        return self.features.num_features if self.has_features else None

    @property
    def features_type(self) -> Optional[str]:
        return self.features.type if self.has_features else None

    @property
    def sampling_rate(self) -> int:
        return self.features.sampling_rate if self.has_features else self.recording.sampling_rate

    def load_features(self) -> Optional[np.ndarray]:
        """
        Load the features from the underlying storage and cut them to the relevant
        [begin, duration] region of the current Cut.
        """
        if self.has_features:
            return self.features.load(start=self.start, duration=self.duration)
        return None

    def load_audio(self) -> Optional[np.ndarray]:
        """
        Load the audio by locating the appropriate recording in the supplied RecordingSet.
        The audio is trimmed to the [begin, end] range specified by the Cut.

        :return: a numpy ndarray with audio samples, with shape (1 <channel>, N <samples>)
        """
        if self.has_recording:
            return self.recording.load_audio(
                channels=self.channel,
                offset_seconds=self.start,
                duration_seconds=self.duration,
            )
        return None

    def compute_and_store_features(
            self,
            extractor: FeatureExtractor,
            storage: FeaturesWriter,
            augmenter: Optional[WavAugmenter] = None,
            *args,
            **kwargs
    ) -> AnyCut:
        """
        Compute the features from this cut, store them on disk, and attach a feature manifest to this cut.
        This cut has to be able to load audio.

        :param extractor: a ``FeatureExtractor`` instance used to compute the features.
        :param output_dir: directory where the computed features will be stored.
        :param augmenter: optional ``WavAugmenter`` instance for audio augmentation.
        :return: a new ``Cut`` instance with a ``Features`` manifest attached to it.
        """
        features_info = extractor.extract_from_samples_and_store(
            samples=self.load_audio(),
            storage=storage,
            sampling_rate=self.sampling_rate,
            offset=self.start,
            augmenter=augmenter,
        )
        # The fastest way to instantiate a copy of the cut with a Features object attached
        return fastcopy(self, features=features_info)

    def truncate(
            self,
            *,
            offset: Seconds = 0.0,
            duration: Optional[Seconds] = None,
            keep_excessive_supervisions: bool = True,
            preserve_id: bool = False
    ) -> 'Cut':
        """
        Returns a new Cut that is a sub-region of the current Cut.

        Note that no operation is done on the actual features - it's only during the call to load_features()
        when the actual changes happen (a subset of features is loaded).

        :param offset: float (seconds), controls the start of the new cut relative to the current Cut's start.
            E.g., if the current Cut starts at 10.0, and offset is 2.0, the new start is 12.0.
        :param duration: optional float (seconds), controls the duration of the resulting Cut.
            By default, the duration is (end of the cut before truncation) - (offset).
        :param keep_excessive_supervisions: bool. Since trimming may happen inside a SupervisionSegment,
            the caller has an option to either keep or discard such supervisions.
        :param preserve_id: bool. Should the truncated cut keep the same ID or get a new, random one.
        :return: a new Cut instance. If the current Cut is shorter than the duration, return None.
        """
        new_start = self.start + offset
        until = offset + (duration if duration is not None else self.duration)
        new_duration = self.duration - new_start if duration is None else until - offset
        assert new_duration > 0.0
        duration_past_end = (new_start + new_duration) - (self.start + self.duration)
        if duration_past_end > 0:
            # When the end of the Cut has been exceeded, trim the new duration to not exceed the old Cut's end.
            new_duration -= duration_past_end
        # Round the duration to 1ms to avoid the possible loss of a single audio sample due to floating point
        # additions and subtractions.
        new_duration = round(new_duration, ndigits=3)
        new_time_span = TimeSpan(start=0, end=new_duration)
        criterion = overlaps if keep_excessive_supervisions else overspans
        new_supervisions = (segment.with_offset(-offset) for segment in self.supervisions)
        return Cut(
            id=self.id if preserve_id else str(uuid4()),
            start=new_start,
            duration=new_duration,
            channel=self.channel,
            supervisions=[
                segment for segment in new_supervisions if criterion(new_time_span, segment)
            ],
            features=self.features,
            recording=self.recording
        )

    def pad(self, duration: Seconds) -> AnyCut:
        """
        Return a new MixedCut, padded to ``duration`` seconds with zeros in the recording,
        and low-energy values in each feature bin.

        :param duration: The cut's minimal duration after padding.
        :return: a padded MixedCut if ``duration`` is greater than this cut's duration, otherwise ``self``.
        """
        if duration <= self.duration:
            return self
        total_num_frames = round(duration / self.frame_shift) if self.has_features else None
        total_num_samples = round(duration * self.sampling_rate) if self.has_recording else None
        padding_duration = round(duration - self.duration, ndigits=3)
        return self.append(PaddingCut(
            id=str(uuid4()),
            duration=padding_duration,
            num_features=self.num_features if self.features is not None else None,
            num_frames=total_num_frames - self.num_frames if self.features is not None else None,
            num_samples=total_num_samples - self.num_samples if self.recording is not None else None,
            sampling_rate=self.features.sampling_rate if self.features is not None else self.recording.sampling_rate,
            use_log_energy=self.features.type in ('fbank', 'mfcc') if self.features is not None else False
        ))

    def map_supervisions(self, transform_fn: Callable[[SupervisionSegment], SupervisionSegment]) -> AnyCut:
        """
        Modify the SupervisionSegments by `transform_fn` of this Cut.

        :param transform_fn: a function that modifies a supervision as an argument.
        :return: a modified Cut.
        """
        new_cut = fastcopy(self, supervisions=[s.map(transform_fn) for s in self.supervisions])
        return new_cut

    @staticmethod
    def from_dict(data: dict) -> 'Cut':
        features = Features.from_dict(data.pop('features')) if 'features' in data else None
        recording = Recording.from_dict(data.pop('recording')) if 'recording' in data else None
        supervision_infos = data.pop('supervisions') if 'supervisions' in data else []
        return Cut(
            **data,
            features=features,
            recording=recording,
            supervisions=[SupervisionSegment.from_dict(s) for s in supervision_infos]
        )

    def with_features_path_prefix(self, path: Pathlike) -> 'Cut':
        if not self.has_features:
            return self
        return fastcopy(self, features=self.features.with_path_prefix(path))

    def with_recording_path_prefix(self, path: Pathlike) -> 'Cut':
        if not self.has_recording:
            return self
        return fastcopy(self, recording=self.recording.with_path_prefix(path))


@dataclass
class PaddingCut(CutUtilsMixin):
    """
    This represents a cut filled with zeroes in the time domain, or low energy/log-energy values in the
    frequency domain. It's used to make training samples evenly sized (same duration/number of frames).
    """
    id: str
    duration: Seconds

    sampling_rate: int
    use_log_energy: bool

    # For frequency domain
    num_frames: Optional[int] = None
    num_features: Optional[int] = None

    # For time domain
    num_samples: Optional[int] = None

    @property
    def supervisions(self):
        return []

    @property
    def has_features(self) -> bool:
        return self.num_frames is not None

    @property
    def has_recording(self) -> bool:
        return self.num_samples is not None

    @property
    def frame_shift(self):
        return round(self.duration / self.num_frames, ndigits=3) if self.has_features else None

    # noinspection PyUnusedLocal
    def load_features(self, *args, **kwargs) -> Optional[np.ndarray]:
        if self.has_features:
            value = np.log(EPSILON) if self.use_log_energy else EPSILON
            return np.ones((self.num_frames, self.num_features), np.float32) * value
        return None

    # noinspection PyUnusedLocal
    def load_audio(self, *args, **kwargs) -> Optional[np.ndarray]:
        if self.has_recording:
            return np.zeros((1, round(self.duration * self.sampling_rate)), np.float32)
        return None

    # noinspection PyUnusedLocal
    def truncate(
            self,
            *,
            offset: Seconds = 0.0,
            duration: Optional[Seconds] = None,
            keep_excessive_supervisions: bool = True,
            preserve_id: bool = False,
    ) -> 'PaddingCut':
        new_duration = self.duration - offset if duration is None else duration
        assert new_duration > 0.0
        return PaddingCut(
            id=self.id if preserve_id else str(uuid4()),
            duration=new_duration,
            num_frames=round(new_duration / self.frame_shift) if self.num_frames is not None else None,
            num_features=self.num_features,
            num_samples=round(new_duration * self.sampling_rate) if self.num_samples is not None else None,
            use_log_energy=self.use_log_energy,
            sampling_rate=self.sampling_rate
        )

    def pad(self, duration: Seconds) -> 'PaddingCut':
        """
        Create a new PaddingCut with ``duration`` when its longer than this Cuts duration.
        Helper function used in batch cut padding.

        :param duration: The cuts minimal duration after padding.
        :return: ``self`` or a new PaddingCut, depending on ``duration``.
        """
        if duration <= self.duration:
            return self
        return PaddingCut(
            id=str(uuid4()),
            duration=duration,
            num_features=self.num_features,
            num_frames=round(duration / self.frame_shift),
            sampling_rate=self.sampling_rate,
            use_log_energy=self.use_log_energy
        )

    def compute_and_store_features(self, extractor: FeatureExtractor, *args, **kwargs) -> AnyCut:
        """
        Returns a new PaddingCut with updates information about the feature dimension and number of
        feature frames, depending on the ``extractor`` properties.
        """
        return fastcopy(
            self,
            num_features=extractor.feature_dim(self.sampling_rate),
            num_frames=round(self.duration / extractor.frame_shift)
        )

    def map_supervisions(self, transform_fn: Callable[[Any], Any]) -> AnyCut:
        """
        Just for consistency with `Cut` and `MixedCut`.

        :param transform_fn: a dummy function that would be never called actually.
        :return: the PaddingCut itself.
        """
        return self

    @staticmethod
    def from_dict(data: dict) -> 'PaddingCut':
        return PaddingCut(**data)

    def with_features_path_prefix(self, path: Pathlike) -> 'PaddingCut':
        return self

    def with_recording_path_prefix(self, path: Pathlike) -> 'PaddingCut':
        return self


@dataclass
class MixTrack:
    """
    Represents a single track in a mix of Cuts. Points to a specific Cut and holds information on
    how to mix it with other Cuts, relative to the first track in a mix.
    """
    cut: Union[Cut, PaddingCut]
    offset: Seconds = 0.0
    snr: Optional[Decibels] = None

    @staticmethod
    def from_dict(data: dict):
        raw_cut = data.pop('cut')
        try:
            cut = Cut.from_dict(raw_cut)
        except TypeError:
            cut = PaddingCut.from_dict(raw_cut)
        return MixTrack(cut, **data)


@dataclass
class MixedCut(CutUtilsMixin):
    """
    Represents a Cut that's created from other Cuts via mix or append operations.
    The actual mixing operations are performed upon loading the features into memory.
    In order to load the features, it needs to access the CutSet object that holds the "ingredient" cuts,
    as it only holds their IDs ("pointers").
    The SNR and offset of all the tracks are specified relative to the first track.
    """
    id: str
    tracks: List[MixTrack]

    @property
    def supervisions(self) -> List[SupervisionSegment]:
        """
        Lists the supervisions of the underlying source cuts.
        Each segment start time will be adjusted by the track offset.
        """
        return [
            segment.with_offset(track.offset)
            for track in self.tracks
            for segment in track.cut.supervisions
        ]

    @property
    def duration(self) -> Seconds:
        track_durations = (track.offset + track.cut.duration for track in self.tracks)
        return round(max(track_durations), ndigits=3)

    @property
    def has_features(self) -> bool:
        return self._first_non_padding_cut.has_features

    @property
    def has_recording(self) -> bool:
        return self._first_non_padding_cut.has_recording

    @property
    def num_frames(self) -> Optional[int]:
        if self.has_features:
            return round(self.duration / self._first_non_padding_cut.frame_shift)
        return None

    @property
    def frame_shift(self) -> Optional[Seconds]:
        return self._first_non_padding_cut.frame_shift

    @property
    def sampling_rate(self) -> Optional[int]:
        return self._first_non_padding_cut.sampling_rate

    @property
    def num_samples(self) -> Optional[int]:
        return round(self.duration * self.sampling_rate)

    @property
    def num_features(self) -> Optional[int]:
        return self._first_non_padding_cut.num_features

    @property
    def features_type(self) -> Optional[str]:
        return self._first_non_padding_cut.features.type if self.has_features else None

    def truncate(
            self,
            *,
            offset: Seconds = 0.0,
            duration: Optional[Seconds] = None,
            keep_excessive_supervisions: bool = True,
            preserve_id: bool = False,
    ) -> 'MixedCut':
        """
        Returns a new MixedCut that is a sub-region of the current MixedCut. This method truncates the underlying Cuts
        and modifies their offsets in the mix, as needed. Tracks that do not fit in the truncated cut are removed.

        Note that no operation is done on the actual features - it's only during the call to load_features()
        when the actual changes happen (a subset of features is loaded).

        :param offset: float (seconds), controls the start of the new cut relative to the current MixedCut's start.
        :param duration: optional float (seconds), controls the duration of the resulting MixedCut.
            By default, the duration is (end of the cut before truncation) - (offset).
        :param keep_excessive_supervisions: bool. Since trimming may happen inside a SupervisionSegment, the caller has
            an option to either keep or discard such supervisions.
        :param preserve_id: bool. Should the truncated cut keep the same ID or get a new, random one.
        :return: a new MixedCut instance.
        """

        new_tracks = []
        old_duration = self.duration
        new_mix_end = old_duration - offset if duration is None else offset + duration

        for track in sorted(self.tracks, key=lambda t: t.offset):
            # First, determine how much of the beginning of the current track we're going to truncate:
            # when the track offset is larger than the truncation offset, we are not truncating the cut;
            # just decreasing the track offset.

            # 'cut_offset' determines how much we're going to truncate the Cut for the current track.
            cut_offset = max(offset - track.offset, 0)
            # 'track_offset' determines the new track's offset after truncation.
            track_offset = max(track.offset - offset, 0)
            # 'track_end' is expressed relative to the beginning of the mix
            # (not to be confused with the 'start' of the underlying Cut)
            track_end = track.offset + track.cut.duration

            if track_end < offset:
                # Omit a Cut that ends before the truncation offset.
                continue

            cut_duration_decrease = 0
            if track_end > new_mix_end:
                if duration is not None:
                    cut_duration_decrease = track_end - new_mix_end
                else:
                    cut_duration_decrease = track_end - old_duration

            # Compute the new Cut's duration after trimming the start and the end.
            new_duration = track.cut.duration - cut_offset - cut_duration_decrease
            if new_duration <= 0:
                # Omit a Cut that is completely outside the time span of the new truncated MixedCut.
                continue

            new_tracks.append(
                MixTrack(
                    cut=track.cut.truncate(
                        offset=cut_offset,
                        duration=new_duration,
                        keep_excessive_supervisions=keep_excessive_supervisions,
                        preserve_id=preserve_id
                    ),
                    offset=track_offset,
                    snr=track.snr
                )
            )
        return MixedCut(id=str(uuid4()), tracks=new_tracks)

    def pad(self, duration: Seconds) -> AnyCut:
        """
        Return a new MixedCut, padded to ``duration`` seconds with zeros in the recording,
        and low-energy values in each feature bin.

        :param duration: The cut's minimal duration after padding.
        :return: a padded MixedCut if duration is greater than this cut's duration, otherwise ``self``.
        """
        if duration <= self.duration:
            return self
        total_num_frames = round(duration / self.frame_shift) if self.has_features else None
        total_num_samples = round(duration * self.sampling_rate) if self.has_recording else None
        padding_duration = round(duration - self.duration, ndigits=3)
        return self.append(PaddingCut(
            id=str(uuid4()),
            duration=padding_duration,
            num_features=self.num_features,
            # The num_frames and sampling_rate fields are tricky, because it is possible to create a MixedCut
            # from Cuts that have different sampling rates and frame shifts. In that case, we are assuming
            # that we should use the values from the reference cut, i.e. the first one in the mix.
            num_frames=(
                total_num_frames - self.num_frames
                if self.has_features
                else None
            ),
            num_samples=(
                total_num_samples - self.num_samples
                if self.has_recording
                else None
            ),
            sampling_rate=self.tracks[0].cut.sampling_rate,
            use_log_energy=self.features_type in ('fbank', 'mfcc')
        ))

    def load_features(self, mixed: bool = True) -> Optional[np.ndarray]:
        """
        Loads the features of the source cuts and mixes them on-the-fly.

        :param mixed: when True (default), returns a 2D array of features mixed in the feature domain.
            Otherwise returns a 3D array with the first dimension equal to the number of tracks.
        :return: A numpy ndarray with features and with shape ``(num_frames, num_features)``,
            or ``(num_tracks, num_frames, num_features)``
        """
        if not self.has_features:
            return None
        first_cut = self.tracks[0].cut
        mixer = FeatureMixer(
            feature_extractor=create_default_feature_extractor(self._first_non_padding_cut.features.type),
            base_feats=first_cut.load_features(),
            frame_shift=first_cut.frame_shift,
        )
        for track in self.tracks[1:]:
            mixer.add_to_mix(
                feats=track.cut.load_features(),
                snr=track.snr,
                offset=track.offset
            )
        if mixed:
            feats = mixer.mixed_feats
            if self.num_frames - feats.shape[0] == 1:
                # Edge case: both offset and cut duration in a given track ended with .5 (e.g. 10.5 and 1.5)
                # and both of them were rounded down, resulting in an off-by-one error.
                # We'll fix the padding here by repeating the last frame.
                # (it's not elegant but I don't have another idea right now...)
                feats = np.vstack([feats, feats[-1, :]])
            return feats
        else:
            return mixer.unmixed_feats

    def load_audio(self, mixed: bool = True) -> Optional[np.ndarray]:
        """
        Loads the audios of the source cuts and mix them on-the-fly.

        :param mixed: When True (default), returns a mono mix of the underlying tracks.
            Otherwise returns a numpy array with the number of channels equal to the number of tracks.
        :return: A numpy ndarray with audio samples and with shape ``(num_channels, num_samples)``
        """
        if not self.has_recording:
            return None
        mixer = AudioMixer(self.tracks[0].cut.load_audio(), sampling_rate=self.tracks[0].cut.sampling_rate)
        for track in self.tracks[1:]:
            mixer.add_to_mix(
                audio=track.cut.load_audio(),
                snr=track.snr,
                offset=track.offset,
            )
        return mixer.mixed_audio if mixed else mixer.unmixed_audio

    def plot_tracks_features(self):
        """
        Display the feature matrix as an image. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt
        fig, axes = plt.subplots(len(self.tracks))
        features = self.load_features(mixed=False)
        fmin, fmax = features.min(), features.max()
        for idx, ax in enumerate(axes):
            ax.imshow(np.flip(features[idx].transpose(1, 0), 0), vmin=fmin, vmax=fmax)
        return axes

    def plot_tracks_audio(self):
        """
        Display plots of the individual tracks' waveforms. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt
        audio = self.load_audio(mixed=False)
        fig, axes = plt.subplots(len(self.tracks), sharex=True, sharey=True)
        for idx, (track, ax) in enumerate(zip(self.tracks, axes)):
            samples = audio[idx, :]
            ax.plot(np.linspace(0, track.offset + track.cut.duration, len(samples)), samples)
            for supervision in track.cut.supervisions:
                supervision = supervision.trim(track.cut.duration)
                ax.axvspan(track.offset + supervision.start, track.offset + supervision.end, color='green', alpha=0.1)
        return axes

    def compute_and_store_features(
            self,
            extractor: FeatureExtractor,
            storage: FeaturesWriter,
            augmenter: Optional[WavAugmenter] = None,
            mix_eagerly: bool = True
    ) -> AnyCut:
        """
        Compute the features from this cut, store them on disk, and create a new `Cut` object with the
        feature manifest attached. This cut has to be able to load audio.

        :param extractor: a ``FeatureExtractor`` instance used to compute the features.
        :param output_dir: directory where the computed features will be stored.
        :param augmenter: optional ``WavAugmenter`` instance for audio augmentation.
        :param mix_eagerly: when False, extract and store the features for each track separately,
            and mix them dynamically when loading the features.
            When True, mix the audio first and store the mixed features, returning a new ``Cut`` instance
            with the same ID. The returned ``Cut`` will not have a ``Recording`` attached.
        :return: a new ``Cut`` instance if ``mix_eagerly`` is True, or returns ``self``
            with each of the tracks containing the ``Features`` manifests.
        """
        if mix_eagerly:
            features_info = extractor.extract_from_samples_and_store(
                samples=self.load_audio(),
                storage=storage,
                sampling_rate=self.sampling_rate,
                offset=0,
                augmenter=augmenter,
            )
            return Cut(
                id=self.id,
                start=0,
                duration=self.duration,
                channel=0,
                supervisions=self.supervisions,
                features=features_info,
                recording=None
            )
        else:  # mix lazily
            new_tracks = [
                MixTrack(
                    cut=track.cut.compute_and_store_features(
                        extractor=extractor,
                        storage=storage,
                        augmenter=augmenter,
                    ),
                    offset=track.offset,
                    snr=track.snr
                )
                for track in self.tracks
            ]
            return MixedCut(id=self.id, tracks=new_tracks)

    def map_supervisions(self, transform_fn: Callable[[SupervisionSegment], SupervisionSegment]) -> AnyCut:
        """
        Modify the SupervisionSegments by `transform_fn` of this MixedCut.

        :param transform_fn: a function that modifies a supervision as an argument.
        :return: a modified MixedCut.
        """
        new_mixed_cut = fastcopy(self)
        for track in new_mixed_cut.tracks:
            track.cut.supervisions = [segment.map(transform_fn) for segment in track.cut.supervisions]
        return new_mixed_cut

    @staticmethod
    def from_dict(data: dict) -> 'MixedCut':
        return MixedCut(id=data['id'], tracks=[MixTrack.from_dict(track) for track in data['tracks']])

    def with_features_path_prefix(self, path: Pathlike) -> 'MixedCut':
        if not self.has_features:
            return self
        return MixedCut(
            id=self.id,
            tracks=[fastcopy(t, cut=t.cut.with_features_path_prefix(path)) for t in self.tracks]
        )

    def with_recording_path_prefix(self, path: Pathlike) -> 'MixedCut':
        if not self.has_recording:
            return self
        return MixedCut(
            id=self.id,
            tracks=[fastcopy(t, cut=t.cut.with_recording_path_prefix(path)) for t in self.tracks]
        )

    @property
    def _first_non_padding_cut(self) -> Cut:
        return [t.cut for t in self.tracks if not isinstance(t.cut, PaddingCut)][0]


@dataclass
class CutSet(JsonMixin, YamlMixin, Sequence[AnyCut]):
    """
    CutSet combines features with their corresponding supervisions.
    It may have wider span than the actual supervisions, provided the features for the whole span exist.
    It is the basic building block of PyTorch-style Datasets for speech/audio processing tasks.
    """
    cuts: Dict[str, AnyCut]

    @property
    def mixed_cuts(self) -> Dict[str, MixedCut]:
        return {id_: cut for id_, cut in self.cuts.items() if isinstance(cut, MixedCut)}

    @property
    def simple_cuts(self) -> Dict[str, Cut]:
        return {id_: cut for id_, cut in self.cuts.items() if isinstance(cut, Cut)}

    @property
    def ids(self) -> Iterable[str]:
        return self.cuts.keys()

    @property
    def speakers(self) -> FrozenSet[str]:
        return frozenset(supervision.speaker for cut in self for supervision in cut.supervisions)

    @staticmethod
    def from_cuts(cuts: Iterable[AnyCut]) -> 'CutSet':
        return CutSet({cut.id: cut for cut in cuts})

    @staticmethod
    def from_manifests(
            recordings: Optional[RecordingSet] = None,
            supervisions: Optional[SupervisionSet] = None,
            features: Optional[FeatureSet] = None
    ) -> 'CutSet':
        """
        Create a CutSet from any combination of supervision, feature and recording manifests.
        At least one of ``recording_set`` or ``feature_set`` is required.
        The Cut boundaries correspond to those found in the ``feature_set``, when available,
        otherwise to those found in the ``recording_set``
        When a ``supervision_set`` is provided, we'll attach to the Cut all supervisions that
        have a matching recording ID and are fully contained in the Cut's boundaries.
        """
        assert features is not None or recordings is not None, \
            "At least one of feature_set and recording_set has to be provided."
        sup_ok, feat_ok, rec_ok = supervisions is not None, features is not None, recordings is not None
        if feat_ok:
            # Case I: Features are provided.
            # Use features to determine the cut boundaries and attach recordings and supervisions as available.
            return CutSet.from_cuts(
                Cut(
                    id=str(uuid4()),
                    start=features.start,
                    duration=features.duration,
                    channel=features.channels,
                    features=features,
                    recording=recordings[features.recording_id] if rec_ok else None,
                    # The supervisions' start times are adjusted if the features object starts at time other than 0s.
                    supervisions=list(supervisions.find(
                        recording_id=features.recording_id,
                        channel=features.channels,
                        start_after=features.start,
                        end_before=features.end,
                        adjust_offset=True
                    )) if sup_ok else []
                )
                for features in features
            )
        # Case II: Recordings are provided (and features are not).
        # Use recordings to determine the cut boundaries.
        return CutSet.from_cuts(
            Cut(
                id=str(uuid4()),
                start=0,
                duration=recording.duration,
                channel=channel,
                recording=recording,
                supervisions=list(supervisions.find(
                    recording_id=recording.id,
                    channel=channel
                )) if sup_ok else []
            )
            for recording in recordings
            # A single cut always represents a single channel. When a recording has multiple channels,
            # we create a new cut for each channel separately.
            for channel in recording.channel_ids
        )

    @staticmethod
    def from_dicts(data: Iterable[dict]) -> 'CutSet':
        def deserialize_one(raw_cut: dict) -> AnyCut:
            cut_type = raw_cut.pop('type')
            if cut_type == 'Cut':
                return Cut.from_dict(raw_cut)
            if cut_type == 'MixedCut':
                return MixedCut.from_dict(raw_cut)
            raise ValueError(f"Unexpected cut type during deserialization: '{cut_type}'")

        return CutSet.from_cuts(deserialize_one(cut) for cut in data)

    def to_dicts(self) -> List[dict]:
        return [{**asdict_nonull(cut), 'type': type(cut).__name__} for cut in self]

    def describe(self) -> None:
        """
        Print a message describing details about the ``CutSet`` - the number of cuts and the
        duration statistics, including the total duration and the percentage of speech segments.

        Example output:
            Cuts count: 547
            Total duration (hours): 326.4
            Speech duration (hours): 79.6 (24.4%)
            ***
            Duration statistics (seconds):
            mean    2148.0
            std      870.9
            min      477.0
            25%     1523.0
            50%     2157.0
            75%     2423.0
            max     5415.0
            dtype: float64
        """
        import pandas as pd
        durations = pd.Series([c.duration for c in self])
        speech_durations = pd.Series([s.trim(c.duration).duration for c in self for s in c.supervisions])
        total_sum = durations.sum()
        speech_sum = speech_durations.sum()
        print('Cuts count:', len(self))
        print(f'Total duration (hours): {total_sum / 3600:.1f}')
        print(f'Speech duration (hours): {speech_sum / 3600:.1f} ({speech_sum / total_sum:.1%})')
        print('***')
        print('Duration statistics (seconds):')
        with pd.option_context('precision', 1):
            print(durations.describe().drop('count'))

    def split(self, num_splits: int, randomize: bool = False) -> List['CutSet']:
        """
        Split the ``CutSet`` into ``num_splits`` pieces of equal size.

        :param num_splits: Requested number of splits.
        :param randomize: Optionally randomize the cuts order first.
        :return: A list of ``CutSet`` pieces.
        """
        return [
            CutSet.from_cuts(subset) for subset in
            split_sequence(self, num_splits=num_splits, randomize=randomize)
        ]

    def filter(self, predicate: Callable[[AnyCut], bool]) -> 'CutSet':
        """
        Return a new CutSet with the Cuts that satisfy the `predicate`.

        :param predicate: a function that takes a cut as an argument and returns bool.
        :return: a filtered CutSet.
        """
        return CutSet.from_cuts(cut for cut in self if predicate(cut))

    def trim_to_supervisions(self) -> 'CutSet':
        """
        Return a new CutSet with Cuts that have identical spans as their supervisions.

        :return: a ``CutSet``.
        """
        return CutSet.from_cuts(
            cut.truncate(offset=segment.start, duration=segment.duration)
            for cut in self
            for segment in cut.supervisions
        )

    def trim_to_unsupervised_segments(self) -> 'CutSet':
        """
        Return a new CutSet with Cuts created from segments that have no supervisions (likely
        silence or noise).

        :return: a ``CutSet``.
        """
        cuts = []
        for cut in self:
            segments = []
            supervisions = sorted(cut.supervisions, key=lambda s: s.start)
            # Check if there is an unsupervised segment at the start of the cut,
            # before the first supervision.
            if supervisions[0].start > 0:
                segments.append((0, supervisions[0].start))
            # Check if there are unsupervised segments between the supervisions.
            for left, right in sliding_window(2, supervisions):
                if overlaps(left, right) or left.end == right.start:
                    continue
                segments.append((left.end, right.start))
            # Check if there is an unsupervised segment after the last supervision,
            # before the cut ends.
            if supervisions[-1].end < cut.duration:
                segments.append((supervisions[-1].end, cut.duration))
            # Create cuts from all found unsupervised segments.
            for start, end in segments:
                cuts.append(cut.truncate(offset=start, duration=end - start))
        return CutSet.from_cuts(cuts)

    def mix_same_recording_channels(self) -> 'CutSet':
        """
        Find cuts that come from the same recording and have matching start and end times, but
        represent different channels. Then, mix them together (in matching groups) and return
        a new ``CutSet`` that contains their mixes. This is useful for processing microphone array
        recordings.

        It is intended to be used as the first operation after creating a new ``CutSet`` (but
        might also work in other circumstances, e.g. if it was cut to windows first).

        Example:
            >>> ami = prepare_ami('path/to/ami')
            >>> cut_set = CutSet.from_manifests(recordings=ami['train']['recordings'])
            >>> multi_channel_cut_set = cut_set.mix_same_recording_channels()

        In the AMI example, the ``multi_channel_cut_set`` will yield MixedCuts that hold all single-channel
        Cuts together.
        """
        if self.mixed_cuts:
            raise ValueError("This operation is not applicable to CutSet's containing MixedCut's.")
        groups = groupby(lambda cut: (cut.recording.id, cut.start, cut.end), self)
        return CutSet.from_cuts(mix_cuts(cuts) for cuts in groups.values())

    def pad(
            self,
            duration: Seconds = None,
    ) -> 'CutSet':
        """
        Return a new CutSet with Cuts padded to ``duration`` in seconds.
        Cuts longer than ``duration`` will not be affected.
        Cuts will be padded to the right (i.e. after the signal).
        :param duration: The cuts minimal duration after padding.
        When not specified, we'll choose the duration of the longest cut in the CutSet.
        :return: A padded CutSet.
        """
        if duration is None:
            duration = max(cut.duration for cut in self)
        return CutSet.from_cuts(cut.pad(duration=duration) for cut in self)

    def truncate(
            self,
            max_duration: Seconds,
            offset_type: str,
            keep_excessive_supervisions: bool = True,
            preserve_id: bool = False
    ) -> 'CutSet':
        """
        Return a new CutSet with the Cuts truncated so that their durations are at most `max_duration`.
        Cuts shorter than `max_duration` will not be changed.
        :param max_duration: float, the maximum duration in seconds of a cut in the resulting manifest.
        :param offset_type: str, can be:
        - 'start' => cuts are truncated from their start;
        - 'end' => cuts are truncated from their end minus max_duration;
        - 'random' => cuts are truncated randomly between their start and their end minus max_duration
        :param keep_excessive_supervisions: bool. When a cut is truncated in the middle of a supervision segment,
        should the supervision be kept.
        :param preserve_id: bool. Should the truncated cut keep the same ID or get a new, random one.
        :return: a new CutSet instance with truncated cuts.
        """
        truncated_cuts = []
        for cut in self:
            if cut.duration <= max_duration:
                truncated_cuts.append(cut)
                continue

            def compute_offset():
                if offset_type == 'start':
                    return 0.0
                last_offset = cut.duration - max_duration
                if offset_type == 'end':
                    return last_offset
                if offset_type == 'random':
                    return random.uniform(0.0, last_offset)
                raise ValueError(f"Unknown 'offset_type' option: {offset_type}")

            truncated_cuts.append(cut.truncate(
                offset=compute_offset(),
                duration=max_duration,
                keep_excessive_supervisions=keep_excessive_supervisions,
                preserve_id=preserve_id
            ))
        return CutSet.from_cuts(truncated_cuts)

    def cut_into_windows(self, duration: Seconds, keep_excessive_supervisions: bool = True) -> 'CutSet':
        """
        Return a new ``CutSet``, made by traversing each ``Cut`` in windows of ``duration`` seconds and
        creating new ``Cut`` out of them.

        The last window might have a shorter duration if there was not enough audio, so you might want to
        use either ``.filter()`` or ``.pad()`` afterwards to obtain a uniform duration ``CutSet``.

        :param duration: Desired duration of the new cuts in seconds.
        :param keep_excessive_supervisions: bool. When a cut is truncated in the middle of a supervision segment,
            should the supervision be kept.
        :return: a new CutSet with cuts made from shorter duration windows.
        """
        new_cuts = []
        for cut in self:
            n_windows = ceil(cut.duration / duration)
            for i in range(n_windows):
                new_cuts.append(cut.truncate(
                    offset=duration * i,
                    duration=duration,
                    keep_excessive_supervisions=keep_excessive_supervisions
                ))
        return CutSet.from_cuts(new_cuts)

    def compute_and_store_features(
            self,
            extractor: FeatureExtractor,
            storage: FeaturesWriter,
            augmenter: Optional[WavAugmenter] = None,
            executor: Optional[Any] = None,
            mix_eagerly: bool = True
    ) -> 'CutSet':
        """
        Modify the current ``CutSet`` with by extracting features and attaching the feature manifests
        to the cuts.

        :param extractor: A ``FeatureExtractor`` instance (either Lhotse's built-in or a custom implementation).
        :param output_dir: Where to store the features.
        :param augmenter: optional ``WavAugmenter`` instance for audio augmentation.
        :param executor: when provided, will be used to parallelize the feature extraction process.
            Any executor satisfying the standard concurrent.futures interface will be suitable;
            e.g. ProcessPoolExecutor, ThreadPoolExecutor, or dask.Client for distributed task
            execution (see: https://docs.dask.org/en/latest/futures.html?highlight=Client#start-dask-client)
        :param mix_eagerly: Related to how the features are extracted for ``MixedCut`` instances, if any are present.
            When False, extract and store the features for each track separately,
            and mix them dynamically when loading the features.
            When True, mix the audio first and store the mixed features, returning a new ``Cut`` instance
            with the same ID. The returned ``Cut`` will not have a ``Recording`` attached.
        :return: a new CutSet instance with the same ``Cut``s, but with attached ``Features`` objects
        """
        if executor is None:
            return CutSet.from_cuts(
                cut.compute_and_store_features(
                    extractor=extractor,
                    storage=storage,
                    augmenter=augmenter,
                    mix_eagerly=mix_eagerly
                )
                for cut in self
            )

        futures = []
        for cut in self:
            futures.append(
                executor.submit(
                    _extract_and_store_features_helper_fn,
                    cut,
                    extractor=extractor,
                    storage=storage,
                    augmenter=augmenter,
                    mix_eagerly=mix_eagerly
                )
            )
        cut_set = CutSet.from_cuts(f.result() for f in futures)
        return cut_set

    def with_features_path_prefix(self, path: Pathlike) -> 'CutSet':
        return CutSet.from_cuts(c.with_features_path_prefix(path) for c in self)

    def with_recording_path_prefix(self, path: Pathlike) -> 'CutSet':
        return CutSet.from_cuts(c.with_recording_path_prefix(path) for c in self)

    def map_supervisions(self, transform_fn: Callable[[SupervisionSegment], SupervisionSegment]) -> 'CutSet':
        """
        Modify the SupervisionSegments by `transform_fn` in this CutSet.

        :param transform_fn: a function that modifies a supervision as an argument.
        :return: a new, modified CutSet.
        """
        return CutSet.from_cuts(cut.map_supervisions(transform_fn) for cut in self)

    def transform_text(self, transform_fn: Callable[[str], str]) -> 'CutSet':
        """
        Return a copy of this ``CutSet`` with all ``SupervisionSegments`` text transformed with ``transform_fn``.
        Useful for text normalization, phonetic transcription, etc.

        :param transform_fn: a function that accepts a string and returns a string.
        :return: a new, modified CutSet.
        """
        return self.map_supervisions(lambda s: s.transform_text(transform_fn))

    def __repr__(self) -> str:
        return f'CutSet(len={len(self)})'

    def __contains__(self, item: Union[str, Cut, MixedCut]) -> bool:
        if isinstance(item, str):
            return item in self.cuts
        else:
            return item.id in self.cuts

    def __getitem__(self, cut_id_or_index: Union[int, str]) -> 'AnyCut':
        if isinstance(cut_id_or_index, str):
            return self.cuts[cut_id_or_index]
        # ~100x faster than list(dict.values())[index] for 100k elements
        return next(val for idx, val in enumerate(self.cuts.values()) if idx == cut_id_or_index)

    def __len__(self) -> int:
        return len(self.cuts)

    def __iter__(self) -> Iterable[AnyCut]:
        return iter(self.cuts.values())

    def __add__(self, other: 'CutSet') -> 'CutSet':
        assert not set(self.cuts.keys()).intersection(other.cuts.keys()), "Conflicting IDs when concatenating CutSets!"
        return CutSet(cuts={**self.cuts, **other.cuts})


def make_windowed_cuts_from_features(
        feature_set: FeatureSet,
        cut_duration: Seconds,
        cut_shift: Optional[Seconds] = None,
        keep_shorter_windows: bool = False
) -> CutSet:
    """
    Converts a FeatureSet to a CutSet by traversing each Features object in - possibly overlapping - windows, and
    creating a Cut out of that area. By default, the last window in traversal will be discarded if it cannot satisfy
    the `cut_duration` requirement.

    :param feature_set: a FeatureSet object.
    :param cut_duration: float, duration of created Cuts in seconds.
    :param cut_shift: optional float, specifies how many seconds are in between the starts of consecutive windows.
        Equals `cut_duration` by default.
    :param keep_shorter_windows: bool, when True, the last window will be used to create a Cut even if
        its duration is shorter than `cut_duration`.
    :return: a CutSet object.
    """
    if cut_shift is None:
        cut_shift = cut_duration
    round_fn = ceil if keep_shorter_windows else floor
    cuts = []
    for features in feature_set:
        # Determine the number of cuts, depending on `keep_shorter_windows` argument.
        # When its true, we'll want to include the residuals in the output; otherwise,
        # we provide only full duration cuts.
        n_cuts = round_fn(features.duration / cut_shift)
        if (n_cuts - 1) * cut_shift + cut_duration > features.duration and not keep_shorter_windows:
            n_cuts -= 1
        for idx in range(n_cuts):
            offset = features.start + idx * cut_shift
            duration = min(cut_duration, features.end - offset)
            cuts.append(
                Cut(
                    id=str(uuid4()),
                    start=offset,
                    duration=duration,
                    channel=features.channels,
                    features=features,
                    supervisions=[]
                )
            )
    return CutSet.from_cuts(cuts)


def mix(
        reference_cut: AnyCut,
        mixed_in_cut: AnyCut,
        offset: Seconds = 0,
        snr: Optional[Decibels] = None
) -> MixedCut:
    """
    Overlay, or mix, two cuts. Optionally the `mixed_in_cut` may be shifted by `offset` seconds
    and scaled down (positive SNR) or scaled up (negative SNR).
    Returns a MixedCut, which contains both cuts and the mix information.
    The actual feature mixing is performed during the call to ``MixedCut.load_features()``.

    :param reference_cut: The reference cut for the mix - offset and snr are specified w.r.t this cut.
    :param mixed_in_cut: The mixed-in cut - it will be offset and rescaled to match the offset and snr parameters.
    :param offset: How many seconds to shift the ``mixed_in_cut`` w.r.t. the ``reference_cut``.
    :param snr: Desired SNR of the `right_cut` w.r.t. the `left_cut` in the mix.
    :return: A MixedCut instance.
    """
    if any(isinstance(cut, PaddingCut) for cut in (reference_cut, mixed_in_cut)) and snr is not None:
        warnings.warn('You are mixing cuts to a padding cut with a specified SNR - '
                      'the resulting energies would be extremely low or high. '
                      'We are setting snr to None, so that the original signal energies will be retained instead.')
        snr = None

    if reference_cut.num_features is not None:
        assert reference_cut.num_features == mixed_in_cut.num_features, "Cannot mix cuts with different feature " \
                                                                        "dimensions. "
    assert offset <= reference_cut.duration, f"Cannot mix cut '{mixed_in_cut.id}' with offset {offset}," \
                                             f" which is greater than cuts {reference_cut.id} duration" \
                                             f" of {reference_cut.duration}"
    # When the left_cut is a MixedCut, take its existing tracks, otherwise create a new track.
    old_tracks = (
        reference_cut.tracks
        if isinstance(reference_cut, MixedCut)
        else [MixTrack(cut=reference_cut)]
    )
    # When the right_cut is a MixedCut, adapt its existing tracks with the new offset and snr,
    # otherwise create a new track.
    new_tracks = (
        [
            MixTrack(
                cut=track.cut,
                offset=round(track.offset + offset, ndigits=3),
                snr=(
                    # When no new SNR is specified, retain whatever was there in the first place.
                    track.snr if snr is None
                    # When new SNR is specified but none was specified before, assign the new SNR value.
                    else snr if track.snr is None
                    # When both new and previous SNR were specified, assign their sum,
                    # as the SNR for each track is defined with regard to the first track energy.
                    else track.snr + snr if snr is not None and track is not None
                    # When no SNR was specified whatsoever, use none.
                    else None
                )
            ) for track in mixed_in_cut.tracks
        ]
        if isinstance(mixed_in_cut, MixedCut)
        else [MixTrack(cut=mixed_in_cut, offset=offset, snr=snr)]
    )
    return MixedCut(
        id=str(uuid4()),
        tracks=old_tracks + new_tracks
    )


def append(
        left_cut: AnyCut,
        right_cut: AnyCut,
        snr: Optional[Decibels] = None
) -> MixedCut:
    """Helper method for functional-style appending of Cuts."""
    return left_cut.append(right_cut, snr=snr)


def mix_cuts(cuts: Iterable[AnyCut]) -> MixedCut:
    """Return a MixedCut that consists of the input Cuts mixed with each other as-is."""
    # The following is a fold (accumulate/aggregate) operation; it starts with cuts[0], and mixes it with cuts[1];
    #  then takes their mix and mixes it with cuts[2]; and so on.
    return reduce(mix, cuts)


def append_cuts(cuts: Iterable[AnyCut]) -> AnyCut:
    """Return a MixedCut that consists of the input Cuts appended to each other as-is."""
    # The following is a fold (accumulate/aggregate) operation; it starts with cuts[0], and appends cuts[1] to it;
    #  then takes their it concatenation and appends cuts[2] to it; and so on.
    return reduce(append, cuts)


def _extract_and_store_features_helper_fn(cut: AnyCut, *args, **kwargs) -> AnyCut:
    return cut.compute_and_store_features(*args, **kwargs)



from contextlib import nullcontext as does_not_raise
from math import isclose

import pytest

from lhotse.cut import CutSet, MixedCut
from lhotse.supervision import SupervisionSegment
from lhotse.test_utils import remove_spaces_from_segment_text


# Note:
# Definitions for `cut1`, `cut2` and `cut_set` parameters are standard Pytest fixtures located in test/cut/conftest.py


def test_append_cut_duration_and_supervisions(cut1, cut2):
    appended_cut = cut1.append(cut2)

    assert isinstance(appended_cut, MixedCut)
    assert appended_cut.duration == 20.0
    assert appended_cut.supervisions == [
        SupervisionSegment(id='sup-1', recording_id='irrelevant', start=0.5, duration=6.0),
        SupervisionSegment(id='sup-2', recording_id='irrelevant', start=7.0, duration=2.0),
        SupervisionSegment(id='sup-3', recording_id='irrelevant', start=13.0, duration=2.5)
    ]


@pytest.mark.parametrize(
    ['offset', 'expected_duration', 'exception_expectation'],
    [
        (0, 10.0, does_not_raise()),
        (1, 11.0, does_not_raise()),
        (5, 15.0, does_not_raise()),
        (10, 20.0, does_not_raise()),
        (100, 'irrelevant', pytest.raises(AssertionError))
    ]
)
def test_overlay_cut_duration_and_supervisions(offset, expected_duration, exception_expectation, cut1, cut2):
    with exception_expectation:
        mixed_cut = cut1.mix(cut2, offset_other_by=offset)

        assert isinstance(mixed_cut, MixedCut)
        assert mixed_cut.duration == expected_duration
        assert mixed_cut.supervisions == [
            SupervisionSegment(id='sup-1', recording_id='irrelevant', start=0.5, duration=6.0),
            SupervisionSegment(id='sup-2', recording_id='irrelevant', start=7.0, duration=2.0),
            SupervisionSegment(id='sup-3', recording_id='irrelevant', start=3.0 + offset, duration=2.5)
        ]


@pytest.fixture
def mixed_feature_cut() -> MixedCut:
    cut_set = CutSet.from_json('test/fixtures/mix_cut_test/overlayed_cut_manifest.json')
    mixed_cut = cut_set['mixed-cut-id']
    assert mixed_cut.num_frames == 1360
    assert isclose(mixed_cut.duration, 13.595)
    return mixed_cut


def test_mixed_cut_load_features_mixed(mixed_feature_cut):
    feats = mixed_feature_cut.load_features()
    assert feats.shape[0] == 1360


def test_mixed_cut_load_features_unmixed(mixed_feature_cut):
    feats = mixed_feature_cut.load_features(mixed=False)
    assert feats.shape[0] == 2
    assert feats.shape[1] == 1360


def test_mixed_cut_map_supervisions(mixed_feature_cut):
    for s in mixed_feature_cut.map_supervisions(remove_spaces_from_segment_text).supervisions:
        if s.text is not None:
            assert ' ' not in s.text


@pytest.fixture
def mixed_audio_cut() -> MixedCut:
    cut_set = CutSet.from_json('test/fixtures/mix_cut_test/overlayed_audio_cut_manifest.json')
    mixed_cut = cut_set['mixed-cut-id']
    assert isclose(mixed_cut.duration, 14.4)
    return mixed_cut


def test_mixed_cut_load_audio_mixed(mixed_audio_cut):
    audio = mixed_audio_cut.load_audio()
    assert audio.shape == (1, 230400)


def test_mixed_cut_load_audio_unmixed(mixed_audio_cut):
    audio = mixed_audio_cut.load_audio(mixed=False)
    assert audio.shape == (2, 230400)



from math import isclose
from typing import Dict

import torch
from torch.utils.data import Dataset

from lhotse.cut import CutSet

EPS = 1e-8


class VadDataset(Dataset):
    """
    The PyTorch Dataset for the voice activity detection task.
    Each item in this dataset is a dict of:

    .. code-block::

        {
            'features': (T x F) tensor
            'is_voice': (T x 1) tensor
        }
    """

    def __init__(
            self,
            cuts: CutSet,
    ):
        super().__init__()
        self.cuts = cuts
        self.cut_ids = list(cuts.ids)

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        cut_id = self.cut_ids[idx]
        cut = self.cuts[cut_id]

        features = torch.from_numpy(cut.load_features())
        assert features.shape[0] == cut.num_frames
        assert isclose(cut.num_frames * cut.frame_shift, cut.duration)

        is_voice = torch.from_numpy(cut.supervisions_feature_mask())

        return {
            'features': features,
            'is_voice': is_voice
        }

    def __len__(self) -> int:
        return len(self.cut_ids)



from dataclasses import dataclass

import numpy as np
from scipy.signal import stft

from lhotse.features import FeatureExtractor
from lhotse.utils import Seconds


@dataclass
class ExampleFeatureExtractorConfig:
    frame_len: Seconds = 0.025
    frame_shift: Seconds = 0.01


# @register_extractor
class ExampleFeatureExtractor(FeatureExtractor):
    """
    A minimal class example, showing how to implement a custom feature extractor in Lhotse.
    Note that the "@register_extractor" decorator is commented out so that this example
    is not visible beyond this file.
    """
    name = 'example-feature-extractor'
    config_type = ExampleFeatureExtractorConfig

    def extract(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        f, t, Zxx = stft(
            samples,
            sampling_rate,
            nperseg=round(self.config.frame_len * sampling_rate),
            noverlap=round(self.frame_shift * sampling_rate)
        )
        # Note: returning a magnitude of the STFT might interact badly with lilcom compression,
        # as it performs quantization of the float values and works best with log-scale quantities.
        # It's advised to turn lilcom compression off, or use log-scale, in such cases.
        return np.abs(Zxx)

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    def feature_dim(self, sampling_rate: int) -> int:
        return (sampling_rate * self.config.frame_len) / 2 + 1



import click

from lhotse.utils import fix_random_seed


@click.group()
@click.option('-s', '--seed', type=int, help='Random seed.')
def cli(seed):
    """
    The shell entry point to Lhotse, a tool and a library for audio data manipulation in high altitudes.
    """
    if seed is not None:
        fix_random_seed(seed)


@cli.group()
def prepare():
    """Command group with data preparation recipes."""
    pass


@cli.group()
def obtain():
    """Command group for download and extract data."""
    pass



from pathlib import Path
from typing import Dict, Optional

import torch
from torch.utils.data import Dataset

from lhotse.cut import CutSet
from lhotse.utils import Pathlike


class DiarizationDataset(Dataset):
    """
    A PyTorch Dataset for the speaker diarization task.
    Our assumptions about speaker diarization are the following:

    * we assume a single channel input (for now), which could be either a true mono signal
        or a beamforming result from a microphone array.
    * we assume that the supervision used for model training is a speech activity matrix, with one
        row dedicated to each speaker (either in the current cut or the whole dataset,
        depending on the settings). The columns correspond to feature frames. Each row is effectively
        a Voice Activity Detection supervision for a single speaker. This setup is somewhat inspired by
        the TS-VAD paper: https://arxiv.org/abs/2005.07272

    Each item in this dataset is a dict of:

    .. code-block::

        {
            'features': (T x F) tensor
            'speaker_activity': (num_speaker x T) tensor
        }

    Constructor arguments:

    :param cuts: a ``CutSet`` used to create the dataset object.
    :param min_speaker_dim: optional int, when specified it will enforce that the matrix shape is at least
        that value (useful for datasets like CHiME 6 where the number of speakers is always 4, but some cuts
        might have less speakers than that).
    :param global_speaker_ids: a bool, indicates whether the same speaker should always retain the same row index
        in the speaker activity matrix (useful for speaker-dependent systems)
    :param root_dir: a prefix path to be attached to the feature files paths.
    """

    def __init__(
            self,
            cuts: CutSet,
            min_speaker_dim: Optional[int] = None,
            global_speaker_ids: bool = False,
    ):
        super().__init__()
        self.cuts = cuts
        self.cut_ids = list(cuts.ids)
        self.speakers = {spk: idx for idx, spk in enumerate(self.cuts.speakers)} if global_speaker_ids else None
        self.min_speaker_dim = min_speaker_dim

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        cut_id = self.cut_ids[idx]
        cut = self.cuts[cut_id]

        features = torch.from_numpy(cut.load_features())
        speaker_activity_matrix = torch.from_numpy(cut.speakers_feature_mask(
            min_speaker_dim=self.min_speaker_dim,
            speaker_to_idx_map=self.speakers
        ))

        return {
            'features': features,
            'speaker_activity': speaker_activity_matrix
        }

    def __len__(self) -> int:
        return len(self.cut_ids)



import click

from lhotse.bin.modes import obtain, prepare
from lhotse.recipes.heroico import download_and_untar, prepare_heroico
from lhotse.utils import Pathlike

__all__ = ['heroico']


@prepare.command(context_settings=dict(show_default=True))
@click.argument('speech_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('transcript_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
def heroico(
        speech_dir: Pathlike,
        transcript_dir: Pathlike,
        output_dir: Pathlike
):
    """heroico Answers ASR data preparation."""
    prepare_heroico(speech_dir, transcript_dir, output_dir)


@obtain.command(context_settings=dict(show_default=True))
@click.argument('target_dir', type=click.Path())
def heroico(
        target_dir: Pathlike
):
    """heroico download."""
    download_and_untar(target_dir)



from dataclasses import dataclass

import numpy as np
import torchaudio

from lhotse.features.base import register_extractor, TorchaudioFeatureExtractor
from lhotse.utils import Seconds, EPSILON


@dataclass
class SpectrogramConfig:
    # Note that `snip_edges` parameter is missing from config: in order to simplify the relationship between
    #  the duration and the number of frames, we are always setting `snip_edges` to False.
    dither: float = 0.0
    window_type: str = "povey"
    # Note that frame_length and frame_shift will be converted to milliseconds before torchaudio/Kaldi sees them
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    remove_dc_offset: bool = True
    round_to_power_of_two: bool = True
    energy_floor: float = 1e-10
    min_duration: float = 0.0
    preemphasis_coefficient: float = 0.97
    raw_energy: bool = True


@register_extractor
class Spectrogram(TorchaudioFeatureExtractor):
    """Log spectrogram feature extractor based on ``torchaudio.compliance.kaldi.spectrogram`` function."""
    name = 'spectrogram'
    config_type = SpectrogramConfig
    feature_fn = staticmethod(torchaudio.compliance.kaldi.spectrogram)

    def feature_dim(self, sampling_rate: int) -> int:
        from torchaudio.compliance.kaldi import _next_power_of_2
        window_size = int(self.config.frame_length * sampling_rate)
        return _next_power_of_2(window_size) // 2 + 1 if self.config.round_to_power_of_two else window_size

    @staticmethod
    def mix(features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float) -> np.ndarray:
        # Torchaudio returns log-power spectrum, hence the need for logsumexp
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b)
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        # Torchaudio returns log-power spectrum, hence the need for exp before the sum
        return float(np.sum(np.exp(features)))



import click

from lhotse.bin.modes import obtain, prepare
from lhotse.recipes.tedlium import download_and_untar, prepare_tedlium
from lhotse.utils import Pathlike


@prepare.command()
@click.argument('tedlium_dir', type=click.Path(exists=True, dir_okay=True, file_okay=False))
@click.argument('output_dir', type=click.Path())
def tedlium(
        tedlium_dir: Pathlike,
        output_dir: Pathlike
):
    """
    TED-LIUM v3 recording and supervision manifest preparation.
    """
    prepare_tedlium(tedlium_root=tedlium_dir, output_dir=output_dir)


@obtain.command()
@click.argument('target_dir', type=click.Path())
def tedlium(
        target_dir: Pathlike
):
    """TED-LIUM v3 download (approx. 11GB)."""
    download_and_untar(target_dir)



import warnings
from typing import Union, List

import numpy as np
import torch


def is_wav_augment_available() -> bool:
    """Returns a boolean indicating if WavAugment is both installed and possible to import."""
    try:
        import augment
        return True
    except:
        return False


class WavAugmenter:
    """
    A wrapper class for WavAugment's effect chain.
    You should construct the ``augment.EffectChain`` beforehand and pass it on to this class.
    For more details on how to augment, see https://github.com/facebookresearch/WavAugment
    """

    def __init__(self, effect_chain, sampling_rate: int):
        # A local import so that ``augment`` can be optional.
        import augment
        self.chain: augment.EffectChain = effect_chain
        self.sampling_rate = sampling_rate

    @staticmethod
    def create_predefined(name: str, sampling_rate: int) -> 'WavAugmenter':
        """
        Create a WavAugmenter class with one of the predefined augmentation setups available in Lhotse.
        Some examples are: "pitch", "reverb", "pitch_reverb_tdrop".

        :param name: the name of the augmentation setup.
        :param sampling_rate: expected sampling rate of the input audio.
        """
        return WavAugmenter(
            effect_chain=_DEFAULT_AUGMENTATIONS[name](sampling_rate=sampling_rate),
            sampling_rate=sampling_rate
        )

    def apply(self, audio: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        """
        Apply the effect chain on the ``audio`` tensor.

        :param audio: a (num_channels, num_samples) shaped tensor placed on the CPU.
        """
        if isinstance(audio, np.ndarray):
            audio = torch.from_numpy(audio)

        assert len(audio.shape) == 2, f"Expected 2-dim tensor with shape (num_channels, num_samples), got {audio.shape}"

        augmented = self.chain.apply(
            input_tensor=audio,
            src_info={
                'channels': audio.shape[0],
                'length': audio.shape[1],
                'rate': self.sampling_rate,
            },
            target_info={
                'channels': 1,
                'length': audio.shape[1],
                'rate': self.sampling_rate,
            }
        )

        # Sox might misbehave sometimes by giving nan/inf if sequences are too short (or silent)
        # and the effect chain includes eg `pitch`
        if torch.isnan(augmented).any() or torch.isinf(augmented).any():
            warnings.warn('NaN/Inf encountered in augmented sox output - returning non-augmented audio.')
            return audio.numpy()
        return augmented.numpy()


_DEFAULT_AUGMENTATIONS = {}


def available_wav_augmentations() -> List[str]:
    """Return a list of augmentation setups available in Lhotse."""
    return list(_DEFAULT_AUGMENTATIONS.keys())


def register_wav_augmentation(fn):
    """
    Register a new augmentation setup that can be retrieved by the name of the function passed in.

    :param fn: a function returning an ``augment.EffectChain`` that performs wav augmentation.
    """
    _DEFAULT_AUGMENTATIONS[fn.__name__] = fn
    return fn


@register_wav_augmentation
def pitch(sampling_rate: int):
    """
    Returns a pitch modification effect for wav augmentation.

    :param sampling_rate: a sampling rate value for which the effect will be created (resampling is needed for pitch).
    """
    import augment
    effect_chain = augment.EffectChain()
    # The pitch effect changes the sampling ratio; we have to compensate for that.
    # Here, we specify 'quick' options on both pitch and rate effects, to speed up things
    effect_chain.pitch("-q", _random_pitch_shift).rate("-q", sampling_rate)
    return effect_chain


@register_wav_augmentation
def reverb(*args, **kwargs):
    """
    Returns a reverb effect for wav augmentation.
    """
    import augment
    effect_chain = augment.EffectChain()
    # Reverb it makes the signal to have two channels,
    # which we combine into 1 by running `channels` w/o parameters
    effect_chain.reverb(50, 50, _random_room_size).channels()
    return effect_chain


@register_wav_augmentation
def pitch_reverb_tdrop(sampling_rate: int):
    """
    Returns an effect chain composed of pitch modification, reverberation and time dropout proposed in:

    * https://github.com/facebookresearch/WavAugment/blob/master/examples/python/librispeech_selfsupervised.py#L152
    * https://arxiv.org/abs/2007.00991

    :param sampling_rate: a sampling rate value for which the effect will be created (resampling is needed for pitch).
    """
    import augment
    effect_chain = augment.EffectChain()
    # The pitch effect changes the sampling ratio; we have to compensate for that.
    # Here, we specify 'quick' options on both pitch and rate effects, to speed up things
    effect_chain.pitch("-q", _random_pitch_shift).rate("-q", sampling_rate)
    # Next effect we add is `reverb`; it adds makes the signal to have two channels,
    # which we combine into 1 by running `channels` w/o parameters
    effect_chain.reverb(50, 50, _random_room_size).channels()
    # Futher, we add an effect that randomly drops one 50ms subsequence
    effect_chain.time_dropout(max_seconds=50 / 1000)
    return effect_chain


def _random_pitch_shift() -> int:
    """The returned values are 1/100ths of a semitone, meaning the default is up to a minor third shift up or down."""
    return np.random.randint(-300, 300)


def _random_room_size() -> int:
    """The returned values correspond to the 'room scale' parameter of SoX."""
    return np.random.randint(0, 100)



from contextlib import nullcontext as does_not_raise

import pytest
from pytest import mark, raises

from lhotse import CutSet
from lhotse.audio import RecordingSet
from lhotse.features import FeatureSet
from lhotse.manipulation import combine, load_manifest
from lhotse.supervision import SupervisionSet
from lhotse.test_utils import DummyManifest


@mark.parametrize('manifest_type', [RecordingSet, SupervisionSet, FeatureSet, CutSet])
def test_split_even(manifest_type):
    manifest = DummyManifest(manifest_type, begin_id=0, end_id=100)
    manifest_subsets = manifest.split(num_splits=2)
    assert len(manifest_subsets) == 2
    assert manifest_subsets[0] == DummyManifest(manifest_type, begin_id=0, end_id=50)
    assert manifest_subsets[1] == DummyManifest(manifest_type, begin_id=50, end_id=100)


@mark.parametrize('manifest_type', [RecordingSet, SupervisionSet, FeatureSet, CutSet])
def test_split_randomize(manifest_type):
    manifest = DummyManifest(manifest_type, begin_id=0, end_id=100)
    manifest_subsets = manifest.split(num_splits=2, randomize=True)
    assert len(manifest_subsets) == 2
    recombined_items = list(manifest_subsets[0]) + list(manifest_subsets[1])
    assert len(recombined_items) == len(manifest)
    # Different ordering (we convert to lists first because the *Set classes might internally
    # re-order after concatenation, e.g. by using dict or post-init sorting)
    assert recombined_items != list(manifest)


@mark.parametrize('manifest_type', [RecordingSet, SupervisionSet, FeatureSet, CutSet])
def test_split_odd(manifest_type):
    manifest = DummyManifest(manifest_type, begin_id=0, end_id=100)
    manifest_subsets = manifest.split(num_splits=3)
    assert len(manifest_subsets) == 3
    assert manifest_subsets[0] == DummyManifest(manifest_type, begin_id=0, end_id=34)
    assert manifest_subsets[1] == DummyManifest(manifest_type, begin_id=34, end_id=68)
    assert manifest_subsets[2] == DummyManifest(manifest_type, begin_id=68, end_id=100)


@mark.parametrize('manifest_type', [RecordingSet, SupervisionSet, FeatureSet, CutSet])
def test_cannot_split_to_more_chunks_than_items(manifest_type):
    manifest = DummyManifest(manifest_type, begin_id=0, end_id=1)
    with pytest.raises(ValueError):
        manifest.split(num_splits=10)


@mark.parametrize('manifest_type', [RecordingSet, SupervisionSet, FeatureSet, CutSet])
def test_combine(manifest_type):
    expected = DummyManifest(manifest_type, begin_id=0, end_id=200)
    combined = combine(
        DummyManifest(manifest_type, begin_id=0, end_id=68),
        DummyManifest(manifest_type, begin_id=68, end_id=136),
        DummyManifest(manifest_type, begin_id=136, end_id=200),
    )
    assert combined == expected


@mark.parametrize(
    ['path', 'exception_expectation'],
    [
        ('test/fixtures/audio.json', does_not_raise()),
        ('test/fixtures/supervision.json', does_not_raise()),
        ('test/fixtures/dummy_feats/feature_manifest.json', does_not_raise()),
        ('test/fixtures/libri/cuts.json', does_not_raise()),
        ('test/fixtures/feature_config.yml', raises(ValueError)),
        ('no/such/path.xd', raises(FileNotFoundError)),
    ]
)
def test_load_any_lhotse_manifest(path, exception_expectation):
    with exception_expectation:
        load_manifest(path)



"""
This is a part of English HUB4 corpus.
It contains Broadcast News data, i.e. audio and transcripts of TV news.
We currently support the following LDC packages:

1997 English Broadcast News Train (HUB4)
  Speech       LDC98S71
  Transcripts  LDC98T28

This data is not available for free - your institution needs to have an LDC subscription.
"""

import re
from itertools import chain
from pathlib import Path
from typing import Dict, Union, Optional, List

from cytoolz import sliding_window

from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, recursion_limit, check_and_rglob

# Since BroadcastNews SGML does not include </time> tags, BeautifulSoup hallucinates them
# in incorrect positions - it nests the <time> segments in each other, making parsing more difficult...
# We are using BS4 for parsing automatically up to <turn> level, and then write a custom parsing function.
EXCLUDE_BEGINNINGS = ['</time', '<overlap', '</overlap']


def prepare_broadcast_news(
        audio_dir: Pathlike,
        transcripts_dir: Pathlike,
        output_dir: Optional[Pathlike] = None,
        absolute_paths: bool = False
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Prepare manifests for 1997 English Broadcast News corpus.
    We create three manifests: one with recordings, one with segments supervisions,
    and one with section supervisions. The latter can be used e.g. for topic segmentation.

    :param audio_dir: Path to ``LDC98S71`` package.
    :param transcripts_dir: Path to ``LDC98T28`` package.
    :param output_dir: Directory where the manifests should be written. Can be omitted to avoid writing.
    :return: A dict with manifests. The keys are: ``{'recordings', 'sections', 'segments'}``.
    """
    audio_paths = check_and_rglob(audio_dir, '*.sph')
    sgml_paths = check_and_rglob(transcripts_dir, '*.sgml')

    recordings = RecordingSet.from_recordings(
        Recording.from_sphere(p, relative_path_depth=None if absolute_paths else 3)
        for p in audio_paths
    )

    # BeautifulSoup has quite inefficient tag-to-string rendering that uses a recursive implementation;
    # on some systems the recursion limit needs to be raised for this to work.
    with recursion_limit(5000):
        supervisions_list = [make_supervisions(p, r) for p, r in zip(sgml_paths, recordings)]
    section_supervisions = SupervisionSet.from_segments(
        chain.from_iterable(sups['sections'] for sups in supervisions_list)
    )
    segment_supervisions = SupervisionSet.from_segments(
        chain.from_iterable(sups['segments'] for sups in supervisions_list)
    )

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        recordings.to_json(output_dir / 'recordings.json')
        section_supervisions.to_json(output_dir / 'sections.json')
        segment_supervisions.to_json(output_dir / 'segments.json')

    return {
        'recordings': recordings,
        'sections': section_supervisions,
        'segments': segment_supervisions
    }


def make_supervisions(sgml_path: Pathlike, recording: Recording) -> Dict[str, List[SupervisionSegment]]:
    """Create supervisions for sections and segments for a given HUB4 recording."""
    doc = try_parse(sgml_path)
    episode = doc.find('episode')
    section_supervisions = []
    text_supervisions = []
    text_idx = 0
    for sec_idx, section in enumerate(doc.find('episode').find_all('section')):
        # Create a "section" supervision segment that informs what's the program and
        # type/topic of a given section.
        # It spans multiple regular segments with spoken content.
        sec_start = float(section.attrs['starttime'])
        section_supervisions.append(SupervisionSegment(
            id=f'{recording.id}_section{sec_idx:03d}',
            recording_id=recording.id,
            start=sec_start,
            duration=round(float(section.attrs['endtime']) - sec_start, ndigits=3),
            channel=0,
            language=episode.attrs['language'],
            custom={
                'section': section.attrs['type'],
                'program': episode.attrs['program']
            }
        ))
        for turn in section.find_all('turn'):
            # An example of the format in each turn:
            #
            # <turn speaker=Peter_Jennings spkrtype=male startTime=336.704 endTime=338.229>
            # <overlap startTime=336.704 endTime=337.575>
            # <time sec=336.704>
            #  time served up until
            # </overlap>
            # <time sec=337.575>
            #  this point?
            # </turn>
            for child in turn.children:
                # Here, we switch to custom parsing code as explained at the top of this script.
                lines = [l for l in str(child).split('\n') if
                         len(l) and not any(l.startswith(b) for b in EXCLUDE_BEGINNINGS)]
                if not lines:
                    continue
                times = []
                texts = []
                for time_marker, text in group_lines_in_time_marker(lines):
                    match = re.search(r'sec="?(\d+\.?\d*)"?', time_marker)
                    times.append(float(match.group(1)))
                    texts.append(text)
                times.append(float(turn.attrs['endtime']))
                # Having parsed the current section into start/end times and text
                # for individual speech segments, create a SupervisionSegment for each one.
                for (start, end), text in zip(sliding_window(2, times), texts):
                    text_supervisions.append(SupervisionSegment(
                        id=f'{recording.id}_segment{text_idx:04d}',
                        recording_id=recording.id,
                        start=start,
                        duration=round(end - start, ndigits=3),
                        channel=0,
                        language=episode.attrs['language'],
                        text=text.strip(),
                        speaker=turn.attrs['speaker'],
                        gender=turn.attrs['spkrtype']
                    ))
                text_idx += 1
    return {
        'sections': section_supervisions,
        'segments': text_supervisions
    }


def try_parse(sgml_path: Path):
    """
    Return a BeautifulSoup object created from an SGML file.
    If it runs into Unicode decoding errors, it will try to determine the file's encoding
    and use iconv to automatically convert it to UTF-8.
    """
    from bs4 import BeautifulSoup
    try:
        return BeautifulSoup(sgml_path.read_text(), 'html.parser')
    except UnicodeDecodeError:
        import subprocess
        from tempfile import NamedTemporaryFile
        encoding = subprocess.check_output(f'file -bi {sgml_path}', shell=True, text=True).split(';')[-1].replace(
            'charset=', '').strip()
        with NamedTemporaryFile() as f:
            subprocess.run(f'iconv -f {encoding} -t utf-8 -o {f.name} {sgml_path}', shell=True, check=True, text=True)
            return BeautifulSoup(f.read(), 'html.parser')


def group_lines_in_time_marker(sgml_lines):
    """This is a helper for the situation when a <time> marker contains multiple lines of text."""
    from itertools import groupby
    # Top-level group allows pairwise iteration with step size of 2, e.g.
    # ['<time...', 'text', '<time...', 'text'] -> [('<time...', 'text'), ('<time...', 'text')]
    return group(
        [
            # Connect multi-lines with a whitespace
            ' '.join(l.strip() for l in lines)
            # groupby() will group the lines depending on whether they start with '<time' or not;
            # is_marker is a bool saying whether the group is a <time> marker,
            # and lines contains the actual lines in that group.
            for is_marker, lines
            in groupby(sgml_lines, key=lambda l: l.startswith('<time'))
        ],
        2
    )


def group(lst, n):
    """group([0,3,4,10,2,3], 2) => [(0,3), (4,10), (2,3)]

    Group a list into consecutive n-tuples. Incomplete tuples are
    discarded e.g.

    Source code by Brian Quinlan from:
    https://code.activestate.com/recipes/303060-group-a-list-into-sequential-n-tuples/

    >>> group(range(10), 3)
    [(0, 1, 2), (3, 4, 5), (6, 7, 8)]
    """
    return zip(*[lst[i::n] for i in range(n)])



from pathlib import Path

import click

from lhotse.bin.modes.cli_base import cli
from lhotse.kaldi import load_kaldi_data_dir
from lhotse.utils import Pathlike

__all__ = ['convert_kaldi']


@cli.command()
@click.argument('data_dir', type=click.Path(exists=True, file_okay=False))
@click.argument('sampling_rate', type=int)
@click.argument('manifest_dir', type=click.Path())
def convert_kaldi(data_dir: Pathlike, sampling_rate: int, manifest_dir: Pathlike):
    """
    Convert a Kaldi data dir DATA_DIR into a directory MANIFEST_DIR of lhotse manifests. Ignores feats.scp.
    The SAMPLING_RATE has to be explicitly specified as it is not available to read from DATA_DIR.
    """
    recording_set, maybe_supervision_set = load_kaldi_data_dir(path=data_dir, sampling_rate=sampling_rate)
    manifest_dir = Path(manifest_dir)
    manifest_dir.mkdir(parents=True, exist_ok=True)
    recording_set.to_json(manifest_dir / 'audio.json')
    if maybe_supervision_set is not None:
        maybe_supervision_set.to_json(manifest_dir / 'supervision.json')



import pytest

from lhotse.audio import RecordingSet, Recording, AudioSource
from lhotse.cut import CutSet
from lhotse.features import FeatureSet, Features
from lhotse.supervision import SupervisionSet, SupervisionSegment


@pytest.fixture()
def libri_recording_set():
    return RecordingSet.from_json('test/fixtures/libri/audio.json')


@pytest.fixture
def libri_cut_set():
    return CutSet.from_json('test/fixtures/libri/cuts.json')


@pytest.fixture
def libri_features_set():
    return FeatureSet.from_json('test/fixtures/libri/feature_manifest.json.gz')


@pytest.fixture
def supervision_set():
    return SupervisionSet.from_json('test/fixtures/supervision.yml')


@pytest.fixture
def libri_cut(libri_cut_set):
    return libri_cut_set['e3e70682-c209-4cac-629f-6fbed82c07cd']


def test_load_none_feats_cut_set():
    cutset = CutSet.from_json('test/fixtures/libri/cuts_no_feats.json')
    cut = list(cutset.cuts.values())[0]
    assert cut.features is None
    assert cut.recording is not None


def test_load_none_recording_cut_set():
    cutset = CutSet.from_json('test/fixtures/libri/cuts_no_recording.json')
    cut = list(cutset.cuts.values())[0]
    assert cut.recording is None
    assert cut.features is not None


def test_load_audio(libri_cut):
    samples = libri_cut.load_audio()
    assert samples.shape[0] == 1  # single channel
    assert samples.shape[1] == 10 * 16000  # samples count = duration * sampling_rate


def test_load_none_audio(libri_cut):
    libri_cut.recording = None
    samples = libri_cut.load_audio()
    assert samples is None


def test_num_frames(libri_cut):
    expected_features_frame_count = round(16.04 / 0.01)  # duration / frame_shift
    assert libri_cut.features.num_frames == expected_features_frame_count

    expected_cut_frame_count = round(10 / 0.01)  # duration / frame_shift
    assert libri_cut.num_frames == expected_cut_frame_count


def test_load_features(libri_cut):
    feats = libri_cut.load_features()
    assert feats.shape[0] == libri_cut.num_frames
    assert feats.shape[1] == libri_cut.features.num_features


def test_load_none_features(libri_cut):
    libri_cut.features = None
    feats = libri_cut.load_features()
    assert feats is None


@pytest.fixture
def dummy_recording_set():
    return RecordingSet.from_recordings([
        Recording(
            id='rec1', sampling_rate=16000, num_samples=160000, duration=10, sources=[
                AudioSource(type='file', channels=[0], source='dummy.wav')
            ]
        )
    ])


@pytest.fixture
def dummy_supervision_set():
    return SupervisionSet.from_segments([
        SupervisionSegment(
            id='sup1', recording_id='rec1', start=3, duration=4, channel=0, text='dummy text'
        )
    ])


@pytest.fixture
def dummy_feature_set():
    return FeatureSet.from_features([
        Features(
            recording_id='rec1', channels=0, start=0, duration=10, type='fbank', num_frames=1000,
            num_features=23, sampling_rate=16000, storage_type='lilcom_files', storage_path='feats',
            storage_key='dummy.llc'
        )
    ])


def test_make_cuts_from_recordings(dummy_recording_set):
    cut_set = CutSet.from_manifests(recordings=dummy_recording_set)
    cut1 = cut_set[0]
    assert cut1.start == 0
    assert cut1.duration == 10.0
    assert cut1.end == 10.0
    assert cut1.channel == 0

    assert len(cut1.supervisions) == 0

    assert cut1.has_recording
    assert cut1.recording == dummy_recording_set.recordings['rec1']
    assert cut1.sampling_rate == 16000
    assert cut1.recording_id == 'rec1'
    assert cut1.num_samples == 160000

    assert not cut1.has_features
    assert cut1.features is None
    assert cut1.frame_shift is None
    assert cut1.num_frames is None
    assert cut1.num_features is None
    assert cut1.features_type is None


def test_make_cuts_from_features(dummy_feature_set):
    cut_set = CutSet.from_manifests(features=dummy_feature_set)
    cut1 = cut_set[0]
    assert cut1.start == 0
    assert cut1.duration == 10.0
    assert cut1.end == 10.0
    assert cut1.channel == 0

    assert len(cut1.supervisions) == 0

    assert not cut1.has_recording
    assert cut1.recording is None
    assert cut1.sampling_rate == 16000
    assert cut1.recording_id == 'rec1'
    assert cut1.num_samples is None

    assert cut1.has_features
    assert cut1.features == dummy_feature_set.features[0]
    assert cut1.frame_shift == 0.01
    assert cut1.num_frames == 1000
    assert cut1.num_features == 23
    assert cut1.features_type == 'fbank'


def test_make_cuts_from_features_recordings(dummy_recording_set, dummy_feature_set):
    cut_set = CutSet.from_manifests(recordings=dummy_recording_set, features=dummy_feature_set)
    cut1 = cut_set[0]
    assert cut1.start == 0
    assert cut1.duration == 10.0
    assert cut1.end == 10.0
    assert cut1.channel == 0

    assert len(cut1.supervisions) == 0

    assert cut1.has_recording
    assert cut1.recording == dummy_recording_set.recordings['rec1']
    assert cut1.sampling_rate == 16000
    assert cut1.recording_id == 'rec1'
    assert cut1.num_samples == 160000

    assert cut1.has_features
    assert cut1.features == dummy_feature_set.features[0]
    assert cut1.frame_shift == 0.01
    assert cut1.num_frames == 1000
    assert cut1.num_features == 23
    assert cut1.features_type == 'fbank'


class TestCutOnSupervisions:
    def test_make_cuts_from_recordings_supervisions(self, dummy_recording_set, dummy_supervision_set):
        cut_set = CutSet.from_manifests(recordings=dummy_recording_set,
                                        supervisions=dummy_supervision_set).trim_to_supervisions()
        cut1 = cut_set[0]
        assert cut1.start == 3.0
        assert cut1.duration == 4.0
        assert cut1.end == 7.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 0.0
        assert cut1.supervisions[0].duration == 4.0
        assert cut1.supervisions[0].end == 4.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert cut1.has_recording
        assert cut1.recording == dummy_recording_set.recordings['rec1']
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples == 16000 * 4

        assert not cut1.has_features
        assert cut1.features is None
        assert cut1.frame_shift is None
        assert cut1.num_frames is None
        assert cut1.num_features is None
        assert cut1.features_type is None

    def test_make_cuts_from_features_supervisions(self, dummy_feature_set, dummy_supervision_set):
        cut_set = CutSet.from_manifests(supervisions=dummy_supervision_set,
                                        features=dummy_feature_set).trim_to_supervisions()
        cut1 = cut_set[0]
        assert cut1.start == 3.0
        assert cut1.duration == 4.0
        assert cut1.end == 7.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 0.0
        assert cut1.supervisions[0].duration == 4.0
        assert cut1.supervisions[0].end == 4.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert not cut1.has_recording
        assert cut1.recording is None
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples is None

        assert cut1.has_features
        assert cut1.features == dummy_feature_set.features[0]
        assert cut1.frame_shift == 0.01
        assert cut1.num_frames == 400
        assert cut1.num_features == 23
        assert cut1.features_type == 'fbank'

    def test_make_cuts_from_recordings_features_supervisions(
            self,
            dummy_recording_set,
            dummy_feature_set,
            dummy_supervision_set
    ):
        cut_set = CutSet.from_manifests(recordings=dummy_recording_set, supervisions=dummy_supervision_set,
                                        features=dummy_feature_set).trim_to_supervisions()
        cut1 = cut_set[0]
        assert cut1.start == 3.0
        assert cut1.duration == 4.0
        assert cut1.end == 7.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 0.0
        assert cut1.supervisions[0].duration == 4.0
        assert cut1.supervisions[0].end == 4.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert cut1.has_recording
        assert cut1.recording == dummy_recording_set.recordings['rec1']
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples == 16000 * 4

        assert cut1.has_features
        assert cut1.features == dummy_feature_set.features[0]
        assert cut1.frame_shift == 0.01
        assert cut1.num_frames == 400
        assert cut1.num_features == 23
        assert cut1.features_type == 'fbank'


class TestNoCutOnSupervisions:
    def test_make_cuts_from_recordings_supervisions(self, dummy_recording_set, dummy_supervision_set):
        cut_set = CutSet.from_manifests(recordings=dummy_recording_set, supervisions=dummy_supervision_set)
        cut1 = cut_set[0]
        assert cut1.start == 0
        assert cut1.duration == 10.0
        assert cut1.end == 10.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 3.0
        assert cut1.supervisions[0].end == 7.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert cut1.has_recording
        assert cut1.recording == dummy_recording_set.recordings['rec1']
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples == 160000

        assert not cut1.has_features
        assert cut1.features is None
        assert cut1.frame_shift is None
        assert cut1.num_frames is None
        assert cut1.num_features is None
        assert cut1.features_type is None

    def test_make_cuts_from_features_supervisions(self, dummy_feature_set, dummy_supervision_set):
        cut_set = CutSet.from_manifests(supervisions=dummy_supervision_set, features=dummy_feature_set)
        cut1 = cut_set[0]
        assert cut1.start == 0
        assert cut1.duration == 10.0
        assert cut1.end == 10.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 3.0
        assert cut1.supervisions[0].end == 7.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert not cut1.has_recording
        assert cut1.recording is None
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples is None

        assert cut1.has_features
        assert cut1.features == dummy_feature_set.features[0]
        assert cut1.frame_shift == 0.01
        assert cut1.num_frames == 1000
        assert cut1.num_features == 23
        assert cut1.features_type == 'fbank'

    def test_make_cuts_from_recordings_features_supervisions(
            self,
            dummy_recording_set,
            dummy_feature_set,
            dummy_supervision_set
    ):
        cut_set = CutSet.from_manifests(recordings=dummy_recording_set, supervisions=dummy_supervision_set,
                                        features=dummy_feature_set)
        cut1 = cut_set[0]
        assert cut1.start == 0
        assert cut1.duration == 10.0
        assert cut1.end == 10.0
        assert cut1.channel == 0

        assert len(cut1.supervisions) == 1
        assert cut1.supervisions[0].id == 'sup1'
        assert cut1.supervisions[0].recording_id == 'rec1'
        assert cut1.supervisions[0].start == 3.0
        assert cut1.supervisions[0].end == 7.0
        assert cut1.supervisions[0].channel == 0
        assert cut1.supervisions[0].text == 'dummy text'

        assert cut1.has_recording
        assert cut1.recording == dummy_recording_set.recordings['rec1']
        assert cut1.sampling_rate == 16000
        assert cut1.recording_id == 'rec1'
        assert cut1.num_samples == 160000

        assert cut1.has_features
        assert cut1.features == dummy_feature_set.features[0]
        assert cut1.frame_shift == 0.01
        assert cut1.num_frames == 1000
        assert cut1.num_features == 23
        assert cut1.features_type == 'fbank'



from typing import Optional

import torch
from torch.utils.data import Dataset

from lhotse.augmentation import WavAugmenter
from lhotse.cut import CutSet
from lhotse.features import FeatureExtractor


class UnsupervisedDataset(Dataset):
    """
    Dataset that contains no supervision - it only provides the features extracted from recordings.
    The returned features are a :class:`torch.Tensor` of shape ``(T x F)``, where T is the number of frames,
    and F is the feature dimension.
    """

    def __init__(self, cuts: CutSet):
        super().__init__()
        self.cuts = cuts
        self.cut_ids = list(cuts.ids)
        self._validate()

    def __getitem__(self, item: int) -> torch.Tensor:
        cut = self.cuts[self.cut_ids[item]]
        feats = cut.load_features()
        return torch.from_numpy(feats)

    def __len__(self):
        return len(self.cuts)

    def _validate(self):
        assert all(cut.has_features for cut in self.cuts)


class UnsupervisedWaveformDataset(UnsupervisedDataset):
    """
    A variant of UnsupervisedDataset that provides waveform samples instead of features.
    The output is a tensor of shape (C, T), with C being the number of channels and T the number of audio samples.
    In this implemenation, there will always be a single channel.
    """

    def __getitem__(self, item: int) -> torch.Tensor:
        cut = self.cuts[self.cut_ids[item]]
        audio = cut.load_audio()
        return torch.from_numpy(audio)

    def _validate(self):
        assert all(cut.has_recording for cut in self.cuts)


class DynamicUnsupervisedDataset(UnsupervisedDataset):
    """
    An example dataset that shows how to use on-the-fly feature extraction in Lhotse.
    It accepts two additional inputs - a FeatureExtractor and an optional WavAugmenter for time-domain data augmentation..
    The output is approximately the same as that of the ``UnsupervisedDataset`` -
    there might be slight differences for ``MixedCut``s, because this dataset mixes them in the time domain,
    and ``UnsupervisedDataset`` does that in the feature domain.
    Cuts that are not mixed will yield identical results in both dataset classes.
    """

    def __init__(
            self,
            feature_extractor: FeatureExtractor,
            cuts: CutSet,
            augmenter: Optional[WavAugmenter] = None,
    ):
        super().__init__(cuts)
        self.feature_extractor = feature_extractor
        self.augmenter = augmenter

    def __getitem__(self, item: int) -> torch.Tensor:
        cut = self.cuts[self.cut_ids[item]]
        features = cut.compute_features(
            extractor=self.feature_extractor,
            augmenter=self.augmenter,
        )
        return torch.from_numpy(features)

    def _validate(self):
        assert all(cut.has_recording for cut in self.cuts)



import pytest

from lhotse.cut import CutSet
from lhotse.dataset.speech_synthesis import SpeechSynthesisDataset


@pytest.fixture
def cut_set():
    return CutSet.from_json('test/fixtures/ljspeech/cuts.json')


def test_speech_synthesis_dataset(cut_set):
    dataset = SpeechSynthesisDataset(cut_set)
    example = dataset[0]
    assert example['audio'].shape[1] > 0
    assert example['features'].shape[0] > 0
    assert len(example['tokens']) > 0



from contextlib import nullcontext as does_not_raise
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional

import numpy as np
import pytest
import torchaudio
from pytest import mark, raises

from lhotse.audio import RecordingSet
from lhotse.augmentation import is_wav_augment_available, WavAugmenter
from lhotse.features import (
    FeatureSet,
    Features,
    FeatureMixer,
    FeatureSetBuilder,
    create_default_feature_extractor,
    Fbank,
    Mfcc,
    Spectrogram, FeatureExtractor
)
from lhotse.features.io import LilcomFilesWriter, LilcomHdf5Writer, NumpyFilesWriter, NumpyHdf5Writer
from lhotse.test_utils import DummyManifest
from lhotse.utils import Seconds, time_diff_to_num_frames

other_params = {}
some_augmentation = None


@mark.parametrize(
    ['feature_type', 'exception_expectation'],
    [
        ('mfcc', does_not_raise()),
        ('fbank', does_not_raise()),
        ('spectrogram', does_not_raise()),
        ('pitch', raises(Exception))
    ]
)
def test_feature_extractor(feature_type, exception_expectation):
    # For now, just test that it runs
    with exception_expectation:
        fe = create_default_feature_extractor(feature_type)
        samples, sr = torchaudio.load('test/fixtures/libri/libri-1088-134315-0000.wav')
        fe.extract(samples=samples, sampling_rate=sr)


def test_feature_extractor_serialization():
    fe = Fbank()
    with NamedTemporaryFile() as f:
        fe.to_yaml(f.name)
        fe_deserialized = Fbank.from_yaml(f.name)
    assert fe_deserialized.config == fe.config


def test_feature_extractor_generic_deserialization():
    fe = Fbank()
    with NamedTemporaryFile() as f:
        fe.to_yaml(f.name)
        fe_deserialized = FeatureExtractor.from_yaml(f.name)
    assert fe_deserialized.config == fe.config


@pytest.mark.parametrize(
    ['format', 'compressed'],
    [
        ('yaml', False),
        ('yaml', True),
        ('json', False),
        ('json', True),
    ]
)
def test_feature_set_serialization(format, compressed):
    feature_set = FeatureSet(
        features=[
            Features(
                recording_id='irrelevant',
                channels=0,
                start=0.0,
                duration=20.0,
                type='fbank',
                num_frames=2000,
                num_features=20,
                sampling_rate=16000,
                storage_type='lilcom',
                storage_path='/irrelevant/',
                storage_key='path.llc'
            )
        ]
    )
    with NamedTemporaryFile(suffix='.gz' if compressed else '') as f:
        if format == 'json':
            feature_set.to_json(f.name)
            feature_set_deserialized = FeatureSet.from_json(f.name)
        if format == 'yaml':
            feature_set.to_yaml(f.name)
            feature_set_deserialized = FeatureSet.from_yaml(f.name)
    assert feature_set_deserialized == feature_set


@mark.parametrize(
    ['recording_id', 'channel', 'start', 'duration', 'exception_expectation', 'expected_num_frames'],
    [
        ('recording-1', 0, 0.0, None, does_not_raise(), 50),  # whole recording
        ('recording-1', 0, 0.0, 0.499, does_not_raise(), 50),  # practically whole recording
        ('recording-2', 0, 0.0, 0.7, does_not_raise(), 70),
        ('recording-2', 0, 0.5, 0.5, does_not_raise(), 50),
        ('recording-2', 1, 0.25, 0.65, does_not_raise(), 65),
        ('recording-nonexistent', 0, 0.0, None, raises(KeyError), None),  # no recording
        ('recording-1', 1000, 0.0, None, raises(KeyError), None),  # no channel
        ('recording-2', 0, 0.5, 1.0, raises(KeyError), None),  # no features between [1.0, 1.5]
        ('recording-2', 0, 1.5, None, raises(KeyError), None),  # no features after 1.0
    ]
)
def test_load_features(
        recording_id: str,
        channel: int,
        start: float,
        duration: float,
        exception_expectation,
        expected_num_frames: Optional[float]
):
    # just test that it loads
    feature_set = FeatureSet.from_json('test/fixtures/dummy_feats/feature_manifest.json')
    with exception_expectation:
        features = feature_set.load(recording_id, channel_id=channel, start=start, duration=duration)
        # expect a matrix
        assert len(features.shape) == 2
        # expect time as the first dimension
        assert features.shape[0] == expected_num_frames


def test_load_features_with_default_arguments():
    feature_set = FeatureSet.from_json('test/fixtures/dummy_feats/feature_manifest.json')
    features = feature_set.load('recording-1')
    assert features.shape == (50, 23)


@pytest.mark.parametrize(
    'storage', [
        LilcomFilesWriter(TemporaryDirectory().name),
        LilcomHdf5Writer(NamedTemporaryFile().name),
        NumpyFilesWriter(TemporaryDirectory().name),
        NumpyHdf5Writer(NamedTemporaryFile().name)
    ]
)
def test_feature_set_builder(storage):
    recordings: RecordingSet = RecordingSet.from_json('test/fixtures/audio.json')
    extractor = Fbank()
    with storage:
        builder = FeatureSetBuilder(
            feature_extractor=extractor,
            storage=storage,
        )
        feature_set = builder.process_and_store_recordings(recordings=recordings)

    assert len(feature_set) == 6

    feature_infos = list(feature_set)

    # Assert the properties shared by all features
    for features in feature_infos:
        # assert that fbank is the default feature type
        assert features.type == 'fbank'
        # assert that duration is always a multiple of frame_shift
        assert features.num_frames == round(features.duration / features.frame_shift)
        # assert that num_features is preserved
        assert features.num_features == builder.feature_extractor.config.num_mel_bins
        # assert that the storage type metadata matches
        assert features.storage_type == storage.name
        # assert that the metadata is consistent with the data shapes
        arr = features.load()
        assert arr.shape[0] == features.num_frames
        assert arr.shape[1] == features.num_features
        # assert that the stored features are the same as the "freshly extracted" features
        recording = recordings[features.recording_id]
        expected = extractor.extract(
            samples=recording.load_audio(channels=features.channels),
            sampling_rate=recording.sampling_rate
        )
        np.testing.assert_almost_equal(arr, expected, decimal=2)

    # Assert the properties for recordings of duration 0.5 seconds
    for features in feature_infos[:2]:
        assert features.num_frames == 50
        assert features.duration == 0.5

    # Assert the properties for recordings of duration 1.0 seconds
    for features in feature_infos[2:]:
        assert features.num_frames == 100
        assert features.duration == 1.0


@pytest.mark.skipif(not is_wav_augment_available(), reason='WavAugment required')
def test_feature_set_builder_with_augmentation():
    recordings: RecordingSet = RecordingSet.from_json('test/fixtures/audio.json')
    augmenter = WavAugmenter.create_predefined('pitch_reverb_tdrop', sampling_rate=8000)
    extractor = Fbank()
    with TemporaryDirectory() as d, LilcomFilesWriter(d) as storage:
        builder = FeatureSetBuilder(
            feature_extractor=extractor,
            storage=storage,
            augmenter=augmenter
        )
        feature_set = builder.process_and_store_recordings(recordings=recordings)

        assert len(feature_set) == 6

        feature_infos = list(feature_set)

        # Assert the properties shared by all features
        for features in feature_infos:
            # assert that fbank is the default feature type
            assert features.type == 'fbank'
            # assert that duration is always a multiple of frame_shift
            assert features.num_frames == round(features.duration / features.frame_shift)
            # assert that num_features is preserved
            assert features.num_features == builder.feature_extractor.config.num_mel_bins
            # assert that the storage type metadata matches
            assert features.storage_type == storage.name
            # assert that the metadata is consistent with the data shapes
            arr = features.load()
            assert arr.shape[0] == features.num_frames
            assert arr.shape[1] == features.num_features

        # Assert the properties for recordings of duration 0.5 seconds
        for features in feature_infos[:2]:
            assert features.num_frames == 50
            assert features.duration == 0.5

        # Assert the properties for recordings of duration 1.0 seconds
        for features in feature_infos[2:]:
            assert features.num_frames == 100
            assert features.duration == 1.0


@mark.parametrize(
    ['time_diff', 'frame_length', 'frame_shift', 'expected_num_frames'],
    [
        (1.0, 0.025, 0.01, 98),
        (0.5, 0.025, 0.01, 48),
        (1.0, 0.025, 0.012, 82),
    ]
)
def test_time_diff_to_num_frames(
        time_diff: Seconds,
        frame_length: Seconds,
        frame_shift: Seconds,
        expected_num_frames: int
):
    assert time_diff_to_num_frames(
        time_diff=time_diff, frame_length=frame_length, frame_shift=frame_shift
    ) == expected_num_frames


def test_add_feature_sets():
    expected = DummyManifest(FeatureSet, begin_id=0, end_id=10)
    feature_set_1 = DummyManifest(FeatureSet, begin_id=0, end_id=5)
    feature_set_2 = DummyManifest(FeatureSet, begin_id=5, end_id=10)
    combined = feature_set_1 + feature_set_2
    assert combined == expected


@pytest.mark.parametrize(
    ['feature_extractor', 'decimal', 'exception_expectation'],
    [
        (Fbank(), 0, does_not_raise()),
        (Spectrogram(), -1, does_not_raise()),
        (Mfcc(), None, raises(ValueError)),
    ]
)
def test_mixer(feature_extractor, decimal, exception_expectation):
    # Treat it more like a test of "it runs" rather than "it works"
    t = np.linspace(0, 1, 8000, dtype=np.float32)
    x1 = np.sin(440.0 * t).reshape(1, -1)
    x2 = np.sin(55.0 * t).reshape(1, -1)

    f1 = feature_extractor.extract(x1, 8000)
    f2 = feature_extractor.extract(x2, 8000)
    with exception_expectation:
        mixer = FeatureMixer(
            feature_extractor=feature_extractor,
            base_feats=f1,
            frame_shift=feature_extractor.frame_shift,
        )
        mixer.add_to_mix(f2)

        fmix_feat = mixer.mixed_feats
        fmix_time = feature_extractor.extract(x1 + x2, 8000)

        np.testing.assert_almost_equal(fmix_feat, fmix_time, decimal=decimal)

        assert mixer.unmixed_feats.shape == (2, 100, feature_extractor.feature_dim(sampling_rate=8000))


def test_feature_set_prefix_path():
    features = FeatureSet.from_features([
        Features(
            type='fbank',
            num_frames=1000,
            num_features=40,
            sampling_rate=16000,
            storage_type='lilcom',
            storage_path='feats/',
            storage_key='12345.llc',
            start=0,
            duration=10
        )
    ])
    for feat in features.with_path_prefix('/data'):
        assert feat.storage_path == '/data/feats'
